"""This module contains the definition of the DatabaseModel class.  The role of this class is to provide an interface
between the BDBag representation of a dataset and a sqlite database in which the contents of the bag are stored.
"""

from __future__ import annotations

import json
import logging
from csv import reader
from pathlib import Path
from typing import Any, Generator, Optional, Type
from urllib.parse import urlparse

from dateutil import parser
from deriva.core.ermrest_model import Column as DerivaColumn
from deriva.core.ermrest_model import Model
from deriva.core.ermrest_model import Table as DerivaTable
from deriva.core.ermrest_model import Type as DerivaType
from pydantic import ConfigDict, validate_call
from sqlalchemy import (
    JSON,
    Boolean,
    Date,
    DateTime,
    Float,
    Integer,
    MetaData,
    String,
    create_engine,
    event,
    inspect,
    select,
)
from sqlalchemy import Column as SQLColumn
from sqlalchemy import ForeignKeyConstraint as SQLForeignKeyConstraint
from sqlalchemy import Table as SQLTable
from sqlalchemy import UniqueConstraint as SQLUniqueConstraint
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import backref, configure_mappers, foreign, relationship
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import TypeDecorator

from deriva_ml.core.definitions import ML_SCHEMA, RID, MLVocab
from deriva_ml.core.exceptions import DerivaMLException
from deriva_ml.dataset.aux_classes import DatasetMinid, DatasetVersion
from deriva_ml.dataset.dataset_bag import DatasetBag
from deriva_ml.model.catalog import DerivaModel

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


class ERMRestBoolean(TypeDecorator):
    impl = Boolean
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value in ("Y", "y", 1, True, "t", "T"):
            return True
        elif value in ("N", "n", 0, False, "f", "F"):
            return False
        elif value is None:
            return None
        raise ValueError(f"Invalid boolean value: {value!r}")


class StringToFloat(TypeDecorator):
    impl = Float
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value == "" or value is None:
            return None
        else:
            return float(value)

class StringToInteger(TypeDecorator):
    impl = Integer
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value == "" or value is None:
            return None
        else:
            return int(value)


class StringToDateTime(TypeDecorator):
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value == "" or value is None:
            return None
        else:
            return parser.parse(value)

class StringToDate(TypeDecorator):
    impl = Date
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value == "" or value is None:
            return None
        else:
            return parser.parse(value).date()

class DatabaseModelMeta(type):
    """Use metaclass to ensure that there is only one instance of a database model per path"""

    _paths_loaded: dict[Path, "DatabaseModel"] = {}

    def __call__(cls, *args, **kwargs):
        logger = logging.getLogger("deriva_ml")
        bag_path: Path = args[1]
        if bag_path.as_posix() not in cls._paths_loaded:
            logger.info(f"Loading {bag_path}")
            cls._paths_loaded[bag_path] = super().__call__(*args, **kwargs)
        return cls._paths_loaded[bag_path]


class DatabaseModel(DerivaModel, metaclass=DatabaseModelMeta):
    """Read in the contents of a BDBag and create a local SQLite database.

        As part of its initialization, this routine will create a sqlite database that has the contents of all the
    tables in the dataset_table.  In addition, any asset tables will the `Filename` column remapped to have the path
    of the local copy of the file. In addition, a local version of the ERMRest model that as used to generate the
    dataset_table is available.

       The sqlite database will not have any foreign key constraints applied, however, foreign-key relationships can be
    found by looking in the ERMRest model.  In addition, as sqlite doesn't support schema, Ermrest schema are added
    to the table name using the convention SchemaName:TableName.  Methods in DatasetBag that have table names as the
    argument will perform the appropriate name mappings.

    Because of nested datasets, it's possible that more than one dataset rid is in a bag, or that a dataset rid might
    appear in more than one database. To help manage this, a global list of all the datasets that have been loaded
    into DatabaseModels, is kept in the class variable `_rid_map`.

    Because you can load different versions of a dataset simultaneously, the dataset RID and version number are tracked,
    and a new sqlite instance is created for every new dataset version present.

    Attributes:
        bag_path (Path): path to the local copy of the BDBag
        minid (DatasetMinid): Minid for the specified bag
        dataset_rid (RID): RID for the specified dataset
        engine (Connection): connection to the sqlalchemy database holding table values
        domain_schema (str): Name of the domain schema
        dataset_table  (Table): the dataset table in the ERMRest model.
    """

    # Maintain a global map of RIDS to versions and databases.
    _rid_map: dict[RID, list[tuple[DatasetVersion, "DatabaseModel"]]] = {}

    def __init__(self, minid: DatasetMinid, bag_path: Path, dbase_path: Path):
        """Create a new DatabaseModel.

        Args:
            minid: Minid for the specified bag.
            bag_path:  Path to the local copy of the BDBag.
        """

        super().__init__(Model.fromfile("file-system", bag_path / "data/schema.json"))

        self.bag_path = bag_path
        self.minid = minid
        self.dataset_rid = minid.dataset_rid
        self.dbase_path = dbase_path / f"{minid.version_rid}"
        self.dbase_path.mkdir(parents=True, exist_ok=True)

        self.engine = create_engine(f"sqlite:///{(self.dbase_path / 'main.db').resolve()}", future=True)
        self.metadata = MetaData()
        self.Base = automap_base(metadata=self.metadata)

        # Attach event listener for *this instance's* engine
        event.listen(self.engine, "connect", self._attach_schemas)

        schema_file = self.bag_path / "data/schema.json"
        with schema_file.open("r") as f:
            self.snaptime = json.load(f)["snaptime"]

        self._logger = logging.getLogger("deriva_ml")
        self._load_model()
        self.ml_schema = ML_SCHEMA
        self._load_database()
        self._logger.info(
            "Creating new database for dataset: %s in %s",
            self.dataset_rid,
            self.dbase_path,
        )
        self.dataset_table = self.model.schemas[self.ml_schema].tables["Dataset"]

        # Now go through the database and pick out all the dataset_table RIDS, along with their versions.
        with self.engine.connect() as conn:
            dataset_version = self.metadata.tables[f"{self.ml_schema}.Dataset_Version"]
            result = conn.execute(select(dataset_version.c.Dataset, dataset_version.c.Version))
            dataset_versions = [t for t in result]

        dataset_versions = [(v[0], DatasetVersion.parse(v[1])) for v in dataset_versions]
        # Get most current version of each rid
        self.bag_rids = {}
        for rid, version in dataset_versions:
            self.bag_rids[rid] = max(self.bag_rids.get(rid, DatasetVersion(0, 1, 0)), version)

        for dataset_rid, dataset_version in self.bag_rids.items():
            version_list = DatabaseModel._rid_map.setdefault(dataset_rid, [])
            version_list.append((dataset_version, self))

    def _attach_schemas(self, dbapi_conn, _conn_record):
        cur = dbapi_conn.cursor()
        for schema in [self.domain_schema, self.ml_schema]:
            schema_file = (self.dbase_path / f"{schema}.db").resolve()
            cur.execute(f"ATTACH DATABASE '{schema_file}' AS '{schema}'")
        cur.close()

    @staticmethod
    def _sql_type(type: DerivaType) -> TypeEngine:
        """Return the SQL type for a Deriva column."""
        return {
            "boolean": ERMRestBoolean,
            "date": StringToDate,
            "float4": StringToFloat,
            "float8": StringToFloat,
            "int2": StringToInteger,
            "int4": StringToInteger,
            "int8": StringToInteger,
            "json": JSON,
            "jsonb": JSON,
            "timestamptz": StringToDateTime,
            "timestamp": StringToDateTime,
        }.get(type.typename, String)

    def _load_model(self) -> None:
        """Create a sqlite database schema that contains all the tables within the catalog from which the BDBag
        was created."""

        def is_key(column: DerivaColumn, table: DerivaTable) -> bool:
            return column in [key.unique_columns[0] for key in table.keys] and column.name == "RID"

        def col(model, name: str):
            # try ORM attribute first
            try:
                return getattr(model, name).property.columns[0]
            except AttributeError:
                # fall back to exact DB column key on the Table
                return model.__table__.c[name]

        def guess_attr_name(col_name: str) -> str:
            return col_name[:-3] if col_name.lower().endswith("_id") else col_name

        database_tables: list[SQLTable] = []
        for schema_name in [self.domain_schema, self.ml_schema]:
            for table in self.model.schemas[schema_name].tables.values():
                database_columns: list[SQLColumn] = []
                for c in table.columns:
                    # clone column (type, nullability, PK, defaults, unique)
                    database_column = SQLColumn(
                        name=c.name,
                        type_=self._sql_type(c.type),  # SQLAlchemy type object is reusable
                        comment=c.comment,
                        default=c.default,
                        primary_key=is_key(c, table),
                        nullable=c.nullok,
                        # NOTE: server_onupdate, computed, etc. can be added if you use them
                    )
                    database_columns.append(database_column)
                database_table = SQLTable(table.name, self.metadata, *database_columns, schema=schema_name)
                for key in table.keys:
                    key_columns = [c.name for c in key.unique_columns]
                    #     if key.name[0] == "RID":
                    #        continue
                    database_table.append_constraint(
                        SQLUniqueConstraint(
                            *key_columns,
                            name=key.name[1],
                        )
                    )
                for fk in table.foreign_keys:
                    if fk.pk_table.schema.name not in [self.domain_schema, self.ml_schema]:
                        continue
                    if fk.pk_table.schema.name != schema_name:
                        continue
                    # Attach FK to the chosen column
                    database_table.append_constraint(
                        SQLForeignKeyConstraint(
                            columns=[f"{c.name}" for c in fk.foreign_key_columns],
                            refcolumns=[f"{schema_name}.{c.table.name}.{c.name}" for c in fk.referenced_columns],
                            name=fk.name[1],
                            comment=fk.comment,
                        )
                    )
                database_tables.append(database_table)
        with self.engine.begin() as conn:
            self.metadata.create_all(conn, tables=database_tables)

        def name_for_scalar_relationship(_base, local_cls, referred_cls, constraint):
            cols = list(constraint.columns) if constraint is not None else []
            if len(cols) == 1:
                name = cols[0].key
                if name in {c.key for c in local_cls.__table__.columns}:
                    name += "_rel"
                return name
            return constraint.name or referred_cls.__name__.lower()

        def name_for_collection_relationship(_base, local_cls, referred_cls, constraint):
            backref_name = constraint.name.replace("_fkey", "_collection")
            return backref_name or (referred_cls.__name__.lower() + "_collection")

        # Now build ORM mappings for the tables.
        self.Base.prepare(
            self.engine,
            name_for_scalar_relationship=name_for_scalar_relationship,
            name_for_collection_relationship=name_for_collection_relationship,
            reflect=True,
        )

        for schema in [self.domain_schema, self.ml_schema]:
            for table in self.model.schemas[schema].tables.values():
                for fk in table.foreign_keys:
                    if fk.pk_table.schema.name not in [self.domain_schema, self.ml_schema]:
                        continue
                    if fk.pk_table.schema.name == schema:
                        continue
                    table_name = f"{schema}.{table.name}"
                    table_class = self.get_orm_class_by_name(table_name)
                    foreign_key_column_name = fk.foreign_key_columns[0].name
                    foreign_key_column = col(table_class, foreign_key_column_name)

                    referenced_table_name = f"{fk.pk_table.schema.name}.{fk.pk_table.name}"
                    referenced_class = self.get_orm_class_by_name(referenced_table_name)
                    referenced_column = col(referenced_class, fk.referenced_columns[0].name)

                    relationship_attr = guess_attr_name(foreign_key_column_name)
                    backref_attr = fk.name[1].replace("_fkey", "_collection")
                    setattr(
                        table_class,
                        relationship_attr,
                        relationship(
                            referenced_class,
                            foreign_keys=[foreign_key_column],
                            primaryjoin=foreign(foreign_key_column) == referenced_column,
                            backref=backref(backref_attr, viewonly=True),
                            viewonly=True,  # set False for write behavior, but best with proper FKs
                        ),
                    )

        # Reflect won't pick up the second FK in the dataset_dataset table.  We need to do it manually
        # dataset_dataset_class = self.get_orm_class_by_name("deriva-ml.Dataset_Dataset")
        # dataset_class = self.get_orm_class_by_name("deriva-ml.Dataset")
        #     dataset_dataset_class.Nested_Dataset = relationship(
        #         dataset_class,
        #         primaryjoin=foreign(dataset_dataset_class.__table__.c["Nested_Dataset"])
        #         == dataset_class.__table__.c["RID"],
        #         foreign_keys=[dataset_dataset_class.__table__.c["Nested_Dataset"]],
        #         backref=backref("nested_dataset_collection", viewonly=True),  # pick a distinct name
        #         viewonly=True,  # keep it read-only unless you truly want writes
        #         overlaps="Dataset,dataset_dataset_collection",  # optional: silence overlap warnings
        #     )
        configure_mappers()

    def _load_database(self) -> None:
        """Load a SQLite database from a bdbag.  THis is done by looking for all the CSV files in the bdbag directory.

        If the file is for an asset table, update the FileName column of the table to have the local file path for
        the materialized file.  Then load into the sqlite database.
        Note: none of the foreign key constraints are included in the database.
        """
        dpath = self.bag_path / "data"
        asset_map = self._localize_asset_table()  # Map of remote to local assets.

        # Find all the CSV files in the subdirectory and load each file into the database.
        for csv_file in Path(dpath).rglob("*.csv"):
            table = csv_file.stem
            schema = self.domain_schema if table in self.model.schemas[self.domain_schema].tables else self.ml_schema
            sql_table = self.metadata.tables[f"{schema}.{table}"]

            with csv_file.open(newline="") as csvfile:
                csv_reader = reader(csvfile)
                column_names = next(csv_reader)

                # Determine which columns in the table has the Filename and the URL
                asset_indexes = (
                    (column_names.index("Filename"), column_names.index("URL")) if self._is_asset(table) else None
                )

                with self.engine.begin() as conn:
                    object_table = [
                        self._localize_asset(o, asset_indexes, asset_map, table == "Dataset")
                        for o in csv_reader
                    ]
                    conn.execute(
                        sqlite_insert(sql_table).on_conflict_do_nothing(),
                        [dict(zip(column_names, row)) for row in object_table],
                    )

    def _localize_asset_table(self) -> dict[str, str]:
        """Use the fetch.txt file in a bdbag to create a map from a URL to a local file path.

        Returns:
            Dictionary that maps a URL to a local file path.

        """
        fetch_map = {}
        try:
            with Path.open(self.bag_path / "fetch.txt", newline="\n") as fetch_file:
                for row in fetch_file:
                    # Rows in fetch.text are tab seperated with URL filename.
                    fields = row.split("\t")
                    local_file = fields[2].replace("\n", "")
                    local_path = f"{self.bag_path}/{local_file}"
                    fetch_map[urlparse(fields[0]).path] = local_path
        except FileNotFoundError:
            dataset_rid = self.bag_path.name.replace("Dataset_", "")
            logging.info(f"No downloaded assets in bag {dataset_rid}")
        return fetch_map

    def _is_asset(self, table_name: str) -> bool:
        """

        Args:
          table_name: str:

        Returns:
            Boolean that is true if the table looks like an asset table.
        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        sname = self.domain_schema if table_name in self.model.schemas[self.domain_schema].tables else self.ml_schema
        asset_table = self.model.schemas[sname].tables[table_name]
        return asset_columns.issubset({c.name for c in asset_table.columns})

    @staticmethod
    def _localize_asset(o: list, indexes: tuple[int, int], asset_map: dict[str, str], debug: bool = False) -> tuple:
        """Given a list of column values for a table, replace the FileName column with the local file name based on
        the URL value.

        Args:
          o: List of values for each column in a table row.
          indexes: A tuple whose first element is the column index of the file name and whose second element
        is the index of the URL in an asset table.  Tuple is None if table is not an asset table.
          o: list:
          indexes: Optional[tuple[int, int]]:

        Returns:
          Tuple of updated column values.

        """
        if indexes:
            file_column, url_column = indexes
            o[file_column] = asset_map[o[url_column]] if o[url_column] else ""
        return tuple(o)

    def find_table(self, table_name: str) -> SQLTable:
        """Find a table in the catalog."""
        # We will look across ml and domain schema to find a table whose name matches.
        table = [t for t in self.metadata.tables if t == table_name or t.split(".")[1] == table_name][0]
        return self.metadata.tables[table]

    def list_tables(self) -> list[str]:
        """List the names of the tables in the catalog

        Returns:
            A list of table names.  These names are all qualified with the Deriva schema name.
        """
        tables = list(self.metadata.tables.keys())
        tables.sort()
        return tables

    def get_dataset(self, dataset_rid: Optional[RID] = None) -> DatasetBag:
        """Get a dataset, or nested dataset from the bag database

        Args:
            dataset_rid: Optional.  If not provided, use the main RID for the bag.  If a value is given, it must
            be the RID for a nested dataset.

        Returns:
            DatasetBag object for the specified dataset.
        """
        if dataset_rid and dataset_rid not in self.bag_rids:
            raise DerivaMLException(f"Dataset RID {dataset_rid} is not in model.")
        return DatasetBag(self, dataset_rid or self.dataset_rid)

    def dataset_version(self, dataset_rid: Optional[RID] = None) -> DatasetVersion:
        """Return the version of the specified dataset."""
        if dataset_rid and dataset_rid not in self.bag_rids:
            DerivaMLException(f"Dataset RID {dataset_rid} is not in model.")
        return self.bag_rids[dataset_rid]

    def find_datasets(self) -> list[dict[str, Any]]:
        """Returns a list of currently available datasets.

        Returns:
             list of currently available datasets.
        """
        atable = next(self.model.schemas[ML_SCHEMA].tables[MLVocab.dataset_type].find_associations()).name

        # Get a list of all the dataset_type values associated with this dataset_table.
        datasets = []
        ds_types = list(self._get_table_contents(atable))
        for dataset in self._get_table_contents("Dataset"):
            my_types = [t for t in ds_types if t["Dataset"] == dataset["RID"]]
            datasets.append(dataset | {MLVocab.dataset_type: [ds[MLVocab.dataset_type] for ds in my_types]})
        return datasets

    def list_dataset_members(self, dataset_rid: RID) -> dict[str, Any]:
        """Returns a list of all the dataset_table entries associated with a dataset."""
        return self.get_dataset(dataset_rid).list_dataset_members()

    def _get_table_contents(self, table: str) -> Generator[dict[str, Any], None, None]:
        """Retrieve the contents of the specified table as a dictionary.

        Args:
            table: Table to retrieve data from. If schema is not provided as part of the table name,
                the method will attempt to locate the schema for the table.

        Returns:
          A generator producing dictionaries containing the contents of the specified table as name/value pairs.
        """

        with self.engine.connect() as conn:
            result = conn.execute(select(self.find_table(table)))
            for row in result.mappings():
                yield dict(row)

    @staticmethod
    def rid_lookup(dataset_rid: RID) -> list[tuple[DatasetVersion, "DatabaseModel"]]:
        """Return a list of DatasetVersion/DatabaseModel instances corresponding to the given RID.

        Args:
            dataset_rid: Rit to be looked up.

        Returns:
            List of DatasetVersion/DatabaseModel instances corresponding to the given RID.

        Raises:
            Raise a DerivaMLException if the given RID is not found.
        """
        try:
            return DatabaseModel._rid_map[dataset_rid]
        except KeyError:
            raise DerivaMLException(f"Dataset {dataset_rid} not found")

    def get_orm_class_by_name(self, table: str) -> Any | None:
        sql_table = self.find_table(table)
        if sql_table is None:
            raise DerivaMLException(f"Table {table} not found")
        return self.get_orm_class_for_table(sql_table)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def get_orm_class_for_table(self, table: SQLTable | DerivaTable | str) -> Any | None:
        if isinstance(table, DerivaTable):
            table = self.metadata.tables[f"{table.schema.name}.{table.name}"]
            if isinstance(table, str):
                table = self.find_table(table)
        for mapper in self.Base.registry.mappers:
            if mapper.persist_selectable is table or table in mapper.tables:
                return mapper.class_
        return None

    @staticmethod
    def _is_association(
        table_class, min_arity=2, max_arity=2, unqualified=True, pure=True, no_overlap=True, return_fkeys=False
    ):
        """Return (truthy) integer arity if self is a matching association, else False.

        min_arity: minimum number of associated fkeys (default 2)
        max_arity: maximum number of associated fkeys (default 2) or None
        unqualified: reject qualified associations when True (default True)
        pure: reject impure associations when True (default True)
        no_overlap: reject overlapping associations when True (default True)
        return_fkeys: return the set of N associated ForeignKeys if True

        The default behavior with no arguments is to test for pure,
        unqualified, non-overlapping, binary associations.

        An association is comprised of several foreign keys which are
        covered by a non-nullable composite row key. This allows
        specific combinations of foreign keys to appear at most once.

        The arity of an association is the number of foreign keys
        being associated. A typical binary association has arity=2.

        An unqualified association contains *only* the foreign key
        material in its row key. Conversely, a qualified association
        mixes in other material which means that a specific
        combination of foreign keys may repeat with different
        qualifiers.

        A pure association contains *only* row key
        material. Conversely, an impure association includes
        additional metadata columns not covered by the row key. Unlike
        qualifiers, impure metadata merely decorates an association
        without augmenting its identifying characteristics.

        A non-overlapping association does not share any columns
        between multiple foreign keys. This means that all
        combinations of foreign keys are possible. Conversely, an
        overlapping association shares some columns between multiple
        foreign keys, potentially limiting the combinations which can
        be represented in an association row.

        These tests ignore the five ERMrest system columns and any
        corresponding constraints.

        """
        if min_arity < 2:
            raise ValueError("An assocation cannot have arity < 2")
        if max_arity is not None and max_arity < min_arity:
            raise ValueError("max_arity cannot be less than min_arity")

        mapper = inspect(table_class).mapper

        # TODO: revisit whether there are any other cases we might
        # care about where system columns are involved?
        non_sys_cols = {col.name for col in mapper.columns if col.name not in {"RID", "RCT", "RMT", "RCB", "RMB"}}
        unique_columns = [
            {c.name for c in constraint.columns}
            for constraint in inspect(table_class).local_table.constraints
            if isinstance(constraint, SQLUniqueConstraint)
        ]

        non_sys_key_colsets = {
            frozenset(unique_column_set)
            for unique_column_set in unique_columns
            if unique_column_set.issubset(non_sys_cols) and len(unique_column_set) > 1
        }

        if not non_sys_key_colsets:
            # reject: not association
            return False

        # choose longest compound key (arbitrary choice with ties!)
        row_key = sorted(non_sys_key_colsets, key=lambda s: len(s), reverse=True)[0]
        foreign_keys = [constraint for constraint in inspect(table_class).relationships.values()]

        covered_fkeys = {fkey for fkey in foreign_keys if {c.name for c in fkey.local_columns}.issubset(row_key)}
        covered_fkey_cols = set()

        if len(covered_fkeys) < min_arity:
            # reject: not enough fkeys in association
            return False
        elif max_arity is not None and len(covered_fkeys) > max_arity:
            # reject: too many fkeys in association
            return False

        for fkey in covered_fkeys:
            fkcols = {c.name for c in fkey.local_columns}
            if no_overlap and fkcols.intersection(covered_fkey_cols):
                # reject: overlapping fkeys in association
                return False
            covered_fkey_cols.update(fkcols)

        if unqualified and row_key.difference(covered_fkey_cols):
            # reject: qualified association
            return False

        if pure and non_sys_cols.difference(row_key):
            # reject: impure association
            return False

        # return (truthy) arity or fkeys
        if return_fkeys:
            return covered_fkeys
        else:
            return len(covered_fkeys)

    def get_orm_association_class(
        self,
        left_cls: Type[Any],
        right_cls: Type[Any],
        min_arity=2,
        max_arity=2,
        unqualified=True,
        pure=True,
        no_overlap=True,
    ):
        """
        Find an association class C by: (1) walking rels on left_cls to a mid class C,
        (2) verifying C also relates to right_cls. Returns (C, C->left, C->right) or None.

        """
        for _, left_rel in inspect(left_cls).relationships.items():
            mid_cls = left_rel.mapper.class_
            is_assoc = self._is_association(mid_cls, return_fkeys=True)
            if not is_assoc:
                continue
            assoc_local_columns_left = list(is_assoc)[0].local_columns
            assoc_local_columns_right = list(is_assoc)[1].local_columns

            found_left = found_right = False
            for r in inspect(left_cls).relationships.values():
                remote_side = list(r.remote_side)[0]
                if remote_side in assoc_local_columns_left:
                    found_left = r
                if remote_side in assoc_local_columns_right:
                    found_left = r
                    # We have left and right backwards from the assocation, so swap them.
                    assoc_local_columns_left, assoc_local_columns_right = (
                        assoc_local_columns_right,
                        assoc_local_columns_left,
                    )
            for r in inspect(right_cls).relationships.values():
                remote_side = list(r.remote_side)[0]
                if remote_side in assoc_local_columns_right:
                    found_right = r
            if found_left != False and found_right != False:
                return mid_cls, found_left.class_attribute, found_right.class_attribute
        return None

    def delete_database(self):
        """

        Args:

        Returns:

        """
        self.dbase_file.unlink()



"""ERMrest data models for DerivaML.

This module provides Pydantic models that represent ERMrest catalog structures. These models are used
throughout DerivaML for defining and manipulating catalog elements like tables, columns, and keys.

Classes:
    FileUploadState: Tracks the state of file uploads.
    VocabularyTerm: Represents terms in controlled vocabularies.
    ColumnDefinition: Defines columns in tables.
    KeyDefinition: Defines primary and unique keys.
    ForeignKeyDefinition: Defines foreign key relationships.
    TableDefinition: Defines complete table structures.
"""

from __future__ import annotations

import warnings
from typing import Any, Iterable

import deriva.core.ermrest_model as em
from deriva.core.ermrest_model import builtin_types
from pydantic import (
    BaseModel,
    Field,
    computed_field,
    field_validator,
    model_serializer,
)

from .constants import RID
from .enums import BuiltinTypes, UploadState

# Pydantic warnings suppression
warnings.filterwarnings("ignore", message='Field name "schema"', category=Warning, module="pydantic")
warnings.filterwarnings(
    "ignore",
    message="fields may not start with an underscore",
    category=Warning,
    module="pydantic",
)


class FileUploadState(BaseModel):
    """Tracks the state and result of a file upload operation.

    Attributes:
        state (UploadState): Current state of the upload (success, failed, etc.).
        status (str): Detailed status message.
        result (Any): Upload result data, if any.
        rid (RID | None): Resource identifier of the uploaded file, if successful.
    """
    state: UploadState
    status: str
    result: Any

    @computed_field
    @property
    def rid(self) -> RID | None:
        return self.result and self.result["RID"]


class VocabularyTerm(BaseModel):
    """Represents a term in a controlled vocabulary.

    A vocabulary term is a standardized entry in a controlled vocabulary table. Each term has
    a primary name, optional synonyms, and identifiers for cross-referencing.

    Attributes:
        name (str): Primary name of the term.
        synonyms (list[str] | None): Alternative names for the term.
        id (str): CURIE (Compact URI) identifier.
        uri (str): Full URI for the term.
        description (str): Explanation of the term's meaning.
        rid (str): Resource identifier in the catalog.

    Example:
        >>> term = VocabularyTerm(
        ...     Name="epithelial",
        ...     Synonyms=["epithelium"],
        ...     ID="tissue:0001",
        ...     URI="http://example.org/tissue/0001",
        ...     Description="Epithelial tissue type",
        ...     RID="1-abc123"
        ... )
    """
    name: str = Field(alias="Name")
    synonyms: list[str] | None = Field(alias="Synonyms")
    id: str = Field(alias="ID")
    uri: str = Field(alias="URI")
    description: str = Field(alias="Description")
    rid: str = Field(alias="RID")

    class Config:
        extra = "ignore"


class ColumnDefinition(BaseModel):
    """Defines a column in an ERMrest table.

    Provides a Pydantic model for defining columns with their types, constraints, and metadata.
    Maps to deriva_py's Column.define functionality.

    Attributes:
        name (str): Name of the column.
        type (BuiltinTypes): ERMrest data type for the column.
        nullok (bool): Whether NULL values are allowed. Defaults to True.
        default (Any): Default value for the column.
        comment (str | None): Description of the column's purpose.
        acls (dict): Access control lists.
        acl_bindings (dict): Dynamic access control bindings.
        annotations (dict): Additional metadata annotations.

    Example:
        >>> col = ColumnDefinition(
        ...     name="score",
        ...     type=BuiltinTypes.float4,
        ...     nullok=False,
        ...     comment="Confidence score between 0 and 1"
        ... )
    """
    name: str
    type: BuiltinTypes
    nullok: bool = True
    default: Any = None
    comment: str | None = None
    acls: dict = Field(default_factory=dict)
    acl_bindings: dict = Field(default_factory=dict)
    annotations: dict = Field(default_factory=dict)

    @field_validator("type", mode="before")
    @classmethod
    def extract_type_name(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return BuiltinTypes(value["typename"])
        else:
            return value

    @model_serializer()
    def serialize_column_definition(self):
        return em.Column.define(
            self.name,
            builtin_types[self.type.value],
            nullok=self.nullok,
            default=self.default,
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations,
        )


class KeyDefinition(BaseModel):
    """Defines a key constraint in an ERMrest table.

    Provides a Pydantic model for defining primary keys and unique constraints.
    Maps to deriva_py's Key.define functionality.

    Attributes:
        colnames (Iterable[str]): Names of columns that form the key.
        constraint_names (Iterable[str]): Names for the key constraints.
        comment (str | None): Description of the key's purpose.
        annotations (dict): Additional metadata annotations.

    Example:
        >>> key = KeyDefinition(
        ...     colnames=["id", "version"],
        ...     constraint_names=["unique_id_version"],
        ...     comment="Unique identifier with version"
        ... )
    """
    colnames: Iterable[str]
    constraint_names: Iterable[str]
    comment: str | None = None
    annotations: dict = Field(default_factory=dict)

    @model_serializer()
    def serialize_key_definition(self):
        return em.Key.define(
            colnames=self.colnames,
            constraint_names=self.constraint_names,
            comment=self.comment,
            annotations=self.annotations,
        )


class ForeignKeyDefinition(BaseModel):
    """Defines a foreign key relationship between tables.

    Provides a Pydantic model for defining foreign key constraints with referential actions
    and metadata. Maps to deriva_py's ForeignKey.define functionality.

    Attributes:
        colnames (Iterable[str]): Names of columns in the referencing table.
        pk_sname (str): Schema name of the referenced table.
        pk_tname (str): Name of the referenced table.
        pk_colnames (Iterable[str]): Names of columns in the referenced table.
        constraint_names (Iterable[str]): Names for the foreign key constraints.
        on_update (str): Action on update of referenced row. Defaults to "NO ACTION".
        on_delete (str): Action on delete of referenced row. Defaults to "NO ACTION".
        comment (str | None): Description of the relationship.
        acls (dict): Access control lists.
        acl_bindings (dict): Dynamic access control bindings.
        annotations (dict): Additional metadata annotations.

    Example:
        >>> fk = ForeignKeyDefinition(
        ...     colnames=["dataset_id"],
        ...     pk_sname="core",
        ...     pk_tname="dataset",
        ...     pk_colnames=["id"],
        ...     on_delete="CASCADE"
        ... )
    """
    colnames: Iterable[str]
    pk_sname: str
    pk_tname: str
    pk_colnames: Iterable[str]
    constraint_names: Iterable[str] = Field(default_factory=list)
    on_update: str = "NO ACTION"
    on_delete: str = "NO ACTION"
    comment: str | None = None
    acls: dict[str, Any] = Field(default_factory=dict)
    acl_bindings: dict[str, Any] = Field(default_factory=dict)
    annotations: dict[str, Any] = Field(default_factory=dict)

    @model_serializer()
    def serialize_fk_definition(self):
        return em.ForeignKey.define(
            fk_colnames=self.colnames,
            pk_sname=self.pk_sname,
            pk_tname=self.pk_tname,
            pk_colnames=self.pk_colnames,
            on_update=self.on_update,
            on_delete=self.on_delete,
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations,
        )


class TableDefinition(BaseModel):
    """Defines a complete table structure in ERMrest.

    Provides a Pydantic model for defining tables with their columns, keys, and relationships.
    Maps to deriva_py's Table.define functionality.

    Attributes:
        name (str): Name of the table.
        column_defs (Iterable[ColumnDefinition]): Column definitions.
        key_defs (Iterable[KeyDefinition]): Key constraint definitions.
        fkey_defs (Iterable[ForeignKeyDefinition]): Foreign key relationship definitions.
        comment (str | None): Description of the table's purpose.
        acls (dict): Access control lists.
        acl_bindings (dict): Dynamic access control bindings.
        annotations (dict): Additional metadata annotations.

    Example:
        >>> table = TableDefinition(
        ...     name="experiment",
        ...     column_defs=[
        ...         ColumnDefinition(name="id", type=BuiltinTypes.text),
        ...         ColumnDefinition(name="date", type=BuiltinTypes.date)
        ...     ],
        ...     comment="Experimental data records"
        ... )
    """
    name: str
    column_defs: Iterable[ColumnDefinition]
    key_defs: Iterable[KeyDefinition] = Field(default_factory=list)
    fkey_defs: Iterable[ForeignKeyDefinition] = Field(default_factory=list)
    comment: str | None = None
    acls: dict = Field(default_factory=dict)
    acl_bindings: dict = Field(default_factory=dict)
    annotations: dict = Field(default_factory=dict)

    @model_serializer()
    def serialize_table_definition(self):
        return em.Table.define(
            tname=self.name,
            column_defs=[c.model_dump() for c in self.column_defs],
            key_defs=[k.model_dump() for k in self.key_defs],
            fkey_defs=[fk.model_dump() for fk in self.fkey_defs],
            comment=self.comment,
            acls=self.acls,
            acl_bindings=self.acl_bindings,
            annotations=self.annotations,
        )



"""
Pytest configuration and shared fixtures.
"""

import os

import pytest

from deriva_ml import DerivaML
from deriva_ml.demo_catalog import (
    create_demo_datasets,
    create_demo_features,
    populate_demo_catalog,
)

from .test_utils import MLCatalog, MLDatasetCatalog, create_jupyter_kernel, destroy_jupyter_kernel


@pytest.fixture(scope="session")
def catalog_host():
    """Get the test host from the environment or use default."""
    return os.environ.get("DERIVA_HOST", "localhost")


@pytest.fixture(scope="session")
def deriva_catalog(catalog_host):
    """Create a demo ML instance for testing with schema, but no data..""" ""
    resource = MLCatalog(catalog_host)
    yield resource
    resource.cleanup()


@pytest.fixture(scope="session")
def populated_catalog(deriva_catalog, tmp_path):
    """Create a demo ML instance for testing with populated domain schema.   Resets after each test."""
    print("Setting up populated catalog for testing... ", end="")

    ml_instance = DerivaML(deriva_catalog.hostname, deriva_catalog.catalog_id, use_minid=False, working_dir=tmp_path)
    populate_demo_catalog(ml_instance)
    return ml_instance


@pytest.fixture(scope="session")
def catalog_with_datasets(deriva_catalog):
    """Create a demo ML instance for testing with populated domain schema.   Resets after each test."""
    print("Setting up catalog with datasets for testing... ", end="")
    return MLDatasetCatalog(deriva_catalog)


@pytest.fixture(scope="function")
def test_ml(deriva_catalog, tmp_path):
    deriva_catalog.reset_demo_catalog()
    """Create a demo ML instance for testing.   Resets after each class.""" ""
    yield DerivaML(deriva_catalog.hostname, deriva_catalog.catalog_id, use_minid=False, working_dir=tmp_path)
    print("Resetting catalog... ", end="")
    deriva_catalog.reset_demo_catalog()


@pytest.fixture(scope="function")
def dataset_test(catalog_with_datasets):
    catalog_with_datasets.reset_catalog()
    return catalog_with_datasets


@pytest.fixture(scope="function")
def notebook_test(deriva_catalog, tmp_path):
    deriva_catalog.reset_demo_catalog()
    create_jupyter_kernel("test_kernel", tmp_path)
    yield DerivaML(deriva_catalog.hostname, deriva_catalog.catalog_id, use_minid=False, working_dir=tmp_path)
    print("Resetting catalog... ", end="")
    deriva_catalog.reset_demo_catalog()
    destroy_jupyter_kernel("test_kernel")


@pytest.fixture(scope="function")
def test_ml_demo_catalog(ml_catalog, tmp_path):
    # reset_demo_catalog(ml_catalog.catalog)
    ml_instance = DerivaML(ml_catalog.hostname, ml_catalog.catalog_id, use_minid=False, working_dir=tmp_path)
    populate_demo_catalog(ml_instance)
    create_demo_features(ml_instance)
    create_demo_datasets(ml_instance)
    return ml_instance


@pytest.fixture(autouse=True)
def log_start():
    print("\n--- Starting test ---")
    yield
    print("\n--- Ending test ---")



import pytest

from deriva_ml import DerivaMLException, DerivaMLInvalidTerm
from deriva_ml.core.definitions import VocabularyTerm


class TestVocabulary:
    def test_vocabulary_create(self, test_ml):
        ml_instance = test_ml
        ml_instance.create_vocabulary("CV1", "A vocab")
        assert next((t for t in ml_instance.model.find_vocabularies() if t.name == "CV1"), None)

        # Check new vocabulary
        assert ml_instance.model.is_vocabulary("CV1")

        # Check for non-vocabulary
        assert not ml_instance.model.is_vocabulary("Dataset")

        # Check for non-existent table
        with pytest.raises(DerivaMLException):
            ml_instance.model.is_vocabulary("FooBar")

        # Check for duplicate
        with pytest.raises(DerivaMLException):
            ml_instance.create_vocabulary("CV1", "A vocab")

    def test_vocabulary_term(self):
        """Test VocabularyTerm model."""
        term = VocabularyTerm(
            Name="Test Term",
            Synonyms=["test", "term"],
            ID="TEST:001",
            URI="http://example.com/test",
            Description="A test term",
            RID="1234",
        )

        assert term.name == "Test Term"
        assert term.synonyms == ["test", "term"]
        assert term.id == "TEST:001"
        assert term.uri == "http://example.com/test"
        assert term.description == "A test term"
        assert term.rid == "1234"

    def test_add_term(self, test_ml):
        ml_instance = test_ml
        ml_instance.create_vocabulary("CV2", "A vocab")

        # Check for new term
        assert len(ml_instance.list_vocabulary_terms("CV2")) == 0
        ml_instance.add_term("CV2", "T1", description="A vocab")
        assert len(ml_instance.list_vocabulary_terms("CV2")) == 1
        assert ml_instance.lookup_term("CV2", "T1").name == "T1"
        with pytest.raises(DerivaMLInvalidTerm):
            ml_instance.lookup_term("CV2", "T2")

        # Check for repeat add
        ml_instance.add_term("CV2", "T1", description="A vocab")
        with pytest.raises(DerivaMLInvalidTerm):
            ml_instance.add_term("CV2", "T1", description="A vocab", exists_ok=False)

    def test_add_term_synonyms(self, test_ml):
        ml_instance = test_ml
        ml_instance.create_vocabulary("CV3", "A vocab")

        ml_instance.add_term("CV3", "T3", synonyms=["S1", "S2"], description="A vocab")
        assert ml_instance.lookup_term("CV3", "S1").name == "T3"
        # Check synonums



"""
File-related utility functions for DerivaML.
"""

from __future__ import annotations

import json
from datetime import date
from pathlib import Path
from socket import gethostname
from typing import Callable, Generator
from urllib.parse import urlparse

import deriva.core.utils.hash_utils as hash_utils
from pydantic import BaseModel, Field, conlist, field_validator, validate_call


class FileSpec(BaseModel):
    """An entry into the File table

    Attributes:
        url: The File url to the url.
        description: The description of the file.
        md5: The MD5 hash of the file.
        length: The length of the file in bytes.
        file_types: A list of file types.  Each files_type should be a defined term in MLVocab.file_type vocabulary.
    """

    url: str = Field(alias="URL", validation_alias="url")
    md5: str = Field(alias="MD5", validation_alias="md5")
    length: int = Field(alias="Length", validation_alias="length")
    description: str | None = Field(default="", alias="Description", validation_alias="description")
    file_types: conlist(str) | None = []

    @field_validator("url")
    @classmethod
    def validate_file_url(cls, url: str) -> str:
        """Examine the provided URL. If it's a local path, convert it into a tag URL.

        Args:
            url: The URL to validate and potentially convert

        Returns:
            The validated/converted URL

        Raises:
            ValidationError: If the URL is not a file URL
        """
        url_parts = urlparse(url)
        if url_parts.scheme == "tag":
            # Already a tag URL, so just return it.
            return url
        elif (not url_parts.scheme) or url_parts.scheme == "file":
            # There is no scheme part of the URL, or it is a file URL, so it is a local file path.
            # Convert to a tag URL.
            return f"tag://{gethostname()},{date.today()}:file://{url_parts.path}"
        else:
            raise ValueError("url is not a file URL")

    @classmethod
    def create_filespecs(
        cls, path: Path | str, description: str, file_types: list[str] | Callable[[Path], list[str]] | None = None
    ) -> Generator[FileSpec, None, None]:
        """Given a file or directory, generate the sequence of corresponding FileSpecs suitable to create a File table.

        Args:
            path: Path to the file or directory.
            description: The description of the file(s)
            file_types: A list of file types or a function that takes a file path and returns a list of file types.

        Returns:
            An iterable of FileSpecs for each file in the directory.
        """

        path = Path(path)
        file_types = file_types or []
        file_types_fn = file_types if callable(file_types) else lambda _x: file_types

        def create_spec(file_path: Path) -> FileSpec:
            hashes = hash_utils.compute_file_hashes(file_path, hashes=frozenset(["md5", "sha256"]))
            md5 = hashes["md5"][0]
            type_list = file_types_fn(file_path)
            return FileSpec(
                length=path.stat().st_size,
                md5=md5,
                description=description,
                url=file_path.as_posix(),
                file_types=type_list if "File" in type_list else ["File"] + type_list,
            )

        files = [path] if path.is_file() else [f for f in Path(path).rglob("*") if f.is_file()]
        return (create_spec(file) for file in files)

    @staticmethod
    def read_filespec(path: Path | str) -> Generator[FileSpec, None, None]:
        """Get FileSpecs from a JSON lines file.

        Args:
         path: Path to the .jsonl file (string or Path).

        Yields:
             A FileSpec object.
        """
        path = Path(path)
        with path.open("r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                yield FileSpec(**json.loads(line))


# Hack round pydantic validate_call and forward reference.
_raw = FileSpec.create_filespecs.__func__
# wrap it with validate_call, then re‐make it a classmethod
FileSpec.create_filespecs = classmethod(validate_call(_raw))



from deriva.core.datapath import Any

from deriva_ml import DatasetSpec
from tests.test_utils import DatasetDescription, DerivaML, MLDatasetCatalog


class TestDataBaseModel:
    def make_frozen(self, e) -> frozenset:
        e.pop("RMT")
        e.pop("RCT")
        if "Filename" in e:
            e.pop("Filename")
        if "Description" in e and not e["Description"]:
            e["Description"] = ""
        return frozenset(e.items())

    def list_datasets(self, dataset_description: DatasetDescription) -> set[str]:
        nested_datasets = {
            ds
            for dset_member in dataset_description.members.get("Dataset", [])
            for ds in self.list_datasets(dset_member)
        }
        return {dataset_description.rid} | nested_datasets

    def compare_catalogs(self, ml_instance: DerivaML, dataset: MLDatasetCatalog, dataset_spec: DatasetSpec):
        reference_datasets = self.list_datasets(dataset.dataset_description)

        snapshot_catalog = DerivaML(
            ml_instance.host_name,
            ml_instance._version_snapshot(dataset_spec),
            working_dir=ml_instance.working_dir,
            use_minid=False,
        )
        bag = ml_instance.download_dataset_bag(dataset_spec)

        pb = snapshot_catalog.pathBuilder
        ds = pb.schemas[snapshot_catalog.ml_schema].tables["Dataset"]
        subject = pb.schemas[snapshot_catalog.domain_schema].tables["Subject"]
        image = pb.schemas[snapshot_catalog.domain_schema].tables["Image"]
        ds_ds = pb.schemas[snapshot_catalog.ml_schema].tables["Dataset_Dataset"]
        ds_subject = pb.schemas[snapshot_catalog.domain_schema].tables["Dataset_Subject"]
        ds_image = pb.schemas[snapshot_catalog.domain_schema].tables["Dataset_Image"]

        # Check to make sure that all of the datasets are present.
        assert {r for r in bag.model.bag_rids.keys()} == {r for r in reference_datasets}
        for dataset_rid in reference_datasets:
            dataset_bag = bag.model.get_dataset(dataset_rid)
            dataset_rids = tuple(ml_instance.list_dataset_children(dataset_rid, recurse=True) + [dataset_rid])
            bag_rids = [b.dataset_rid for b in dataset_bag.list_dataset_children(recurse=True)] + [
                dataset_bag.dataset_rid
            ]
            assert list(dataset_rids).sort() == bag_rids.sort()

            ds_path = ds.path.link(ds_ds).filter(ds_ds.Dataset == Any(*dataset_rids)).link(ds)
            subject_path = ds.path.link(ds_subject).filter(ds_subject.Dataset == Any(*dataset_rids)).link(subject)
            image_path = ds.path.link(ds_image).filter(ds_image.Dataset == Any(*dataset_rids)).link(image)
            subject_path_1 = (
                ds.path.link(ds_image).filter(ds_image.Dataset == Any(*dataset_rids)).link(image).link(subject)
            )
            image_path_1 = (
                ds.path.link(ds_subject).filter(ds_subject.Dataset == Any(*dataset_rids)).link(subject).link(image)
            )
            dataset_path_1 = ds.path.filter(ds.RID == Any(*dataset_rids))
            datasets = list(ds_path.entities().fetch()) + list(dataset_path_1.entities().fetch())
            subjects = list(subject_path.entities().fetch()) + list(subject_path_1.entities().fetch())
            images = list(image_path.entities().fetch()) + list(image_path_1.entities().fetch())

            catalog_dataset = set([self.make_frozen(d) for d in datasets])
            catalog_subject = set([self.make_frozen(s) for s in subjects])
            catalog_image = set([self.make_frozen(i) for i in images])

            dataset_table = set([self.make_frozen(d) for d in dataset_bag.get_table_as_dict("Dataset")])
            subject_table = set([self.make_frozen(s) for s in dataset_bag.get_table_as_dict("Subject")])
            image_table = set([self.make_frozen(i) for i in dataset_bag.get_table_as_dict("Image")])

            assert len(catalog_dataset) == len(dataset_table)
            assert len(catalog_subject) == len(subject_table)
            assert len(catalog_image) == len(image_table)

            assert catalog_dataset == dataset_table
            assert catalog_subject == subject_table
            assert catalog_image == image_table

    def test_database_methods(self, dataset_test):
        ml_instance = DerivaML(dataset_test.catalog.hostname, dataset_test.catalog.catalog_id, use_minid=False)
        dataset_description = dataset_test.dataset_description
        current_version = ml_instance.dataset_version(dataset_description.rid)
        current_spec = DatasetSpec(rid=dataset_description.rid, version=current_version)
        current_bag = ml_instance.download_dataset_bag(current_spec)
        tables = current_bag.model.list_tables()
        schemas = ml_instance.model.schemas
        catalog_tables = len(schemas[ml_instance.domain_schema].tables) + len(schemas[ml_instance.ml_schema].tables)
        assert catalog_tables == len(tables)

    def test_table_as_dict(self, dataset_test):
        ml_instance = DerivaML(dataset_test.catalog.hostname, dataset_test.catalog.catalog_id, use_minid=False)
        dataset_description = dataset_test.dataset_description

        current_version = ml_instance.dataset_version(dataset_description.rid)
        dataset_spec = DatasetSpec(rid=dataset_description.rid, version=current_version)
        self.compare_catalogs(ml_instance, dataset_test, dataset_spec)

    def test_table_versions(self, dataset_test):
        ml_instance = DerivaML(dataset_test.catalog.hostname, dataset_test.catalog.catalog_id, use_minid=False)
        dataset_description = dataset_test.dataset_description

        current_version = ml_instance.dataset_version(dataset_description.rid)
        current_spec = DatasetSpec(rid=dataset_description.rid, version=current_version)
        self.compare_catalogs(
            ml_instance, dataset_test, DatasetSpec(rid=dataset_description.rid, version=current_version)
        )

        pb = ml_instance.pathBuilder
        subject = pb.schemas[ml_instance.domain_schema].Subject
        new_subjects = [s["RID"] for s in subject.insert([{"Name": f"Mew Thing{t + 1}"} for t in range(2)])]
        ml_instance.add_dataset_members(dataset_description.rid, new_subjects)
        print("Adding subjects: ", new_subjects)
        new_version = ml_instance.dataset_version(dataset_description.rid)
        new_spec = DatasetSpec(rid=dataset_description.rid, version=new_version)
        current_bag = ml_instance.download_dataset_bag(current_spec)
        new_bag = ml_instance.download_dataset_bag(new_spec)
        subjects_current = list(current_bag.get_table_as_dict("Subject"))
        subjects_new = list(new_bag.get_table_as_dict("Subject"))

        # Make sure that there is a difference between to old and new catalogs.
        print(new_subjects)
        print([s["RID"] for s in subjects_current])
        print([s["RID"] for s in subjects_new])
        assert len(subjects_new) == len(subjects_current) + 2
        print("compare")
        self.compare_catalogs(ml_instance, dataset_test, current_spec)
        self.compare_catalogs(ml_instance, dataset_test, new_spec)



"""
Model management for Deriva ML catalogs.

This module provides the DerivaModel class which augments the standard Deriva model class with
ML-specific functionality. It handles schema management, feature definitions, and asset tracking.
"""

from __future__ import annotations

# Standard library imports
from collections import Counter, defaultdict
from graphlib import CycleError, TopologicalSorter
from typing import Any, Callable, Final, Iterable, NewType, TypeAlias

from deriva.core.ermrest_catalog import ErmrestCatalog

# Deriva imports
from deriva.core.ermrest_model import Column, FindAssociationResult, Model, Schema, Table

# Third-party imports
from pydantic import ConfigDict, validate_call

from deriva_ml.core.definitions import (
    ML_SCHEMA,
    RID,
    DerivaAssetColumns,
    TableDefinition,
)
from deriva_ml.core.exceptions import DerivaMLException, DerivaMLTableTypeError

# Local imports
from deriva_ml.feature import Feature
from deriva_ml.protocols.dataset import DatasetLike

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


# Define common types:
TableInput: TypeAlias = str | Table
SchemaDict: TypeAlias = dict[str, Schema]
FeatureList: TypeAlias = Iterable[Feature]
SchemaName = NewType("SchemaName", str)
ColumnSet: TypeAlias = set[Column]
AssociationResult: TypeAlias = FindAssociationResult
TableSet: TypeAlias = set[Table]
PathList: TypeAlias = list[list[Table]]

# Define constants:
VOCAB_COLUMNS: Final[set[str]] = {"NAME", "URI", "SYNONYMS", "DESCRIPTION", "ID"}
ASSET_COLUMNS: Final[set[str]] = {"Filename", "URL", "Length", "MD5", "Description"}

FilterPredicate = Callable[[Table], bool]


class DerivaModel:
    """Augmented interface to deriva model class.

    This class provides a number of DerivaML specific methods that augment the interface in the deriva model class.

    Attributes:
        domain_schema: Schema name for domain-specific tables and relationships.
        model: ERMRest model for the catalog.
        catalog: ERMRest catalog for the model
        hostname: ERMRest catalog for the model
        ml_schema: The ML schema for the catalog.
        domain_schema: The domain schema for the catalog.

    """

    def __init__(
        self,
        model: Model,
        ml_schema: str = ML_SCHEMA,
        domain_schema: str | None = None,
    ):
        """Create and initialize a DerivaML instance.

        This method will connect to a catalog, and initialize local configuration for the ML execution.
        This class is intended to be used as a base class on which domain-specific interfaces are built.

        Args:
            model: The ERMRest model for the catalog.
            ml_schema: The ML schema name.
            domain_schema: The domain schema name.
        """
        self.model = model
        self.configuration = None
        self.catalog: ErmrestCatalog = self.model.catalog
        self.hostname = self.catalog.deriva_server.server if isinstance(self.catalog, ErmrestCatalog) else "localhost"

        self.ml_schema = ml_schema
        builtin_schemas = ("public", self.ml_schema, "www", "WWW")
        if domain_schema:
            self.domain_schema = domain_schema
        else:
            if len(user_schemas := {k for k in self.model.schemas.keys()} - set(builtin_schemas)) == 1:
                self.domain_schema = user_schemas.pop()
            else:
                raise DerivaMLException(f"Ambiguous domain schema: {user_schemas}")

    def refresh_model(self) -> None:
        self.model = self.catalog.getCatalogModel()

    @property
    def schemas(self) -> dict[str, Schema]:
        return self.model.schemas

    @property
    def chaise_config(self) -> dict[str, Any]:
        """Return the chaise configuration."""
        return self.model.chaise_config

    def __getattr__(self, name: str) -> Any:
        # Called only if `name` is not found in Manager.  Delegate attributes to model class.
        return getattr(self.model, name)

    def name_to_table(self, table: TableInput) -> Table:
        """Return the table object corresponding to the given table name.

        If the table name appears in more than one schema, return the first one you find.

        Args:
          table: A ERMRest table object or a string that is the name of the table.

        Returns:
          Table object.
        """
        if isinstance(table, Table):
            return table
        if table in (s := self.model.schemas[self.domain_schema].tables):
            return s[table]
        for s in [self.model.schemas[sname] for sname in [self.domain_schema, self.ml_schema, "WWW"]]:
            if table in s.tables.keys():
                return s.tables[table]
        raise DerivaMLException(f"The table {table} doesn't exist.")

    def is_vocabulary(self, table_name: TableInput) -> bool:
        """Check if a given table is a controlled vocabulary table.

        Args:
          table_name: A ERMRest table object or the name of the table.

        Returns:
          Table object if the table is a controlled vocabulary, False otherwise.

        Raises:
          DerivaMLException: if the table doesn't exist.

        """
        vocab_columns = {"NAME", "URI", "SYNONYMS", "DESCRIPTION", "ID"}
        table = self.name_to_table(table_name)
        return vocab_columns.issubset({c.name.upper() for c in table.columns})

    def is_association(
        self,
        table_name: str | Table,
        unqualified: bool = True,
        pure: bool = True,
        min_arity: int = 2,
        max_arity: int = 2,
    ) -> bool | set[str] | int:
        """Check the specified table to see if it is an association table.

        Args:
            table_name: param unqualified:
            pure: return: (Default value = True)
            table_name: str | Table:
            unqualified:  (Default value = True)

        Returns:


        """
        table = self.name_to_table(table_name)
        return table.is_association(unqualified=unqualified, pure=pure, min_arity=min_arity, max_arity=max_arity)

    def find_association(self, table1: Table | str, table2: Table | str) -> tuple[Table, Column, Column]:
        """Given two tables, return an association table that connects the two and the two columns used to link them..

        Raises:
            DerivaML exception if there is either not an association table or more than one association table.
        """
        table1 = self.name_to_table(table1)
        table2 = self.name_to_table(table2)

        tables = [
            (a.table, a.self_fkey.columns[0].name, other_key.columns[0].name)
            for a in table1.find_associations(pure=False)
            if len(a.other_fkeys) == 1 and (other_key := a.other_fkeys.pop()).pk_table == table2
        ]

        if len(tables) == 1:
            return tables[0]
        elif len(tables) == 0:
            raise DerivaMLException(f"No association tables found between {table1.name} and {table2.name}.")
        else:
            raise DerivaMLException(
                f"There are {len(tables)} association tables between {table1.name} and {table2.name}."
            )

    def is_asset(self, table_name: TableInput) -> bool:
        """True if the specified table is an asset table.

        Args:
            table_name: str | Table:

        Returns:
            True if the specified table is an asset table, False otherwise.

        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        table = self.name_to_table(table_name)
        return asset_columns.issubset({c.name for c in table.columns})

    def find_assets(self, with_metadata: bool = False) -> list[Table]:
        """Return the list of asset tables in the current model"""
        return [t for s in self.model.schemas.values() for t in s.tables.values() if self.is_asset(t)]

    def find_vocabularies(self) -> list[Table]:
        """Return a list of all the controlled vocabulary tables in the domain schema."""
        return [t for s in self.model.schemas.values() for t in s.tables.values() if self.is_vocabulary(t)]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def find_features(self, table: TableInput) -> Iterable[Feature]:
        """List the names of the features in the specified table.

        Args:
            table: The table to find features for.
            table: Table | str:

        Returns:
            An iterable of FeatureResult instances that describe the current features in the table.
        """
        table = self.name_to_table(table)

        def is_feature(a: FindAssociationResult) -> bool:
            """Check if association represents a feature.

            Args:
                a: Association result to check
            Returns:
                bool: True if association represents a feature
            """
            return {
                "Feature_Name",
                "Execution",
                a.self_fkey.foreign_key_columns[0].name,
            }.issubset({c.name for c in a.table.columns})

        return [
            Feature(a, self) for a in table.find_associations(min_arity=3, max_arity=3, pure=False) if is_feature(a)
        ]

    def lookup_feature(self, table: TableInput, feature_name: str) -> Feature:
        """Lookup the named feature associated with the provided table.

        Args:
            table: param feature_name:
            table: str | Table:
            feature_name: str:

        Returns:
            A Feature class that represents the requested feature.

        Raises:
          DerivaMLException: If the feature cannot be found.
        """
        table = self.name_to_table(table)
        try:
            return [f for f in self.find_features(table) if f.feature_name == feature_name][0]
        except IndexError:
            raise DerivaMLException(f"Feature {table.name}:{feature_name} doesn't exist.")

    def asset_metadata(self, table: str | Table) -> set[str]:
        """Return the metadata columns for an asset table."""

        table = self.name_to_table(table)

        if not self.is_asset(table):
            raise DerivaMLTableTypeError("asset table", table.name)
        return {c.name for c in table.columns} - DerivaAssetColumns

    def apply(self) -> None:
        """Call ERMRestModel.apply"""
        if self.catalog == "file-system":
            raise DerivaMLException("Cannot apply() to non-catalog model.")
        else:
            self.model.apply()

    def list_dataset_element_types(self) -> list[Table]:
        """
        Lists the data types of elements contained within a dataset.

        This method analyzes the dataset and identifies the data types for all
        elements within it. It is useful for understanding the structure and
        content of the dataset and allows for better manipulation and usage of its
        data.

        Returns:
            list[str]: A list of strings where each string represents a data type
            of an element found in the dataset.

        """

        dataset_table = self.name_to_table("Dataset")

        def domain_table(table: Table) -> bool:
            return table.schema.name == self.domain_schema or table.name == dataset_table.name

        return [t for a in dataset_table.find_associations() if domain_table(t := a.other_fkeys.pop().pk_table)]

    def _prepare_wide_table(self,
                            dataset,
                            dataset_rid: RID,
                            include_tables: list[str]) -> tuple[dict[str, Any], list[tuple]]:
        """
        Generates details of a wide table from the model

        Args:
            include_tables (list[str] | None): List of table names to include in the denormalized dataset. If None,
                all tables from the dataset will be included.

        Returns:
            str: SQL query string that represents the process of denormalization.
        """

        # Skip over tables that we don't want to include in the denormalized dataset.
        # Also, strip off the Dataset/Dataset_X part of the path so we don't include dataset columns in the denormalized
        # table.
        include_tables = set(include_tables)
        for t in include_tables:
            # Check to make sure the table is in the catalog.
            _ = self.name_to_table(t)

        table_paths = [
            path
            for path in self._schema_to_paths()
            if path[-1].name in include_tables and include_tables.intersection({p.name for p in path})
        ]
        paths_by_element = defaultdict(list)
        for p in table_paths:
            paths_by_element[p[2].name].append(p)

        # Get the names of all of the tables that can be dataset elements.
        dataset_element_tables = {
            e.name for e in self.list_dataset_element_types() if e.schema.name == self.domain_schema
        }

        skip_columns = {"RCT", "RMT", "RCB", "RMB"}
        element_tables = {}
        for element_table, paths in paths_by_element.items():
            graph = {}
            for path in paths:
                for left, right in zip(path[0:], path[1:]):
                    graph.setdefault(left.name, set()).add(right.name)

            # New lets remove any cycles that we may have in the graph.
            # We will use a topological sort to find the order in which we need to join the tables.
            # If we find a cycle, we will remove the table from the graph and splice in an additional ON clause.
            # We will then repeat the process until there are no cycles.
            graph_has_cycles = True
            element_join_tables = []
            element_join_conditions = {}
            while graph_has_cycles:
                try:
                    ts = TopologicalSorter(graph)
                    element_join_tables = list(reversed(list(ts.static_order())))
                    graph_has_cycles = False
                except CycleError as e:
                    cycle_nodes = e.args[1]
                    if len(cycle_nodes) > 3:
                        raise DerivaMLException(f"Unexpected cycle found when normalizing dataset {cycle_nodes}")
                    # Remove cycle from graph and splice in additional ON constraint.
                    graph[cycle_nodes[1]].remove(cycle_nodes[0])

            # The Dataset_Version table is a special case as it points to dataset and dataset to version.
            if "Dataset_Version" in element_join_tables:
                element_join_tables.remove("Dataset_Version")

            for path in paths:
                for left, right in zip(path[0:], path[1:]):
                    if right.name == "Dataset_Version":
                        # The Dataset_Version table is a special case as it points to dataset and dataset to version.
                        continue
                    if element_join_tables.index(right.name) < element_join_tables.index(left.name):
                        continue
                    table_relationship = self._table_relationship(left, right)
                    element_join_conditions.setdefault(right.name, set()).add(
                        (table_relationship[0], table_relationship[1])
                    )
            element_tables[element_table] = (element_join_tables, element_join_conditions)
        # Get the list of columns that will appear in the final denormalized dataset.
        denormalized_columns = [
            (table_name, c.name)
            for table_name in include_tables
            if not self.is_association(table_name)  # Don't include association columns in the denormalized view.'
            for c in self.name_to_table(table_name).columns
            if (not include_tables or table_name in include_tables) and (c.name not in skip_columns)
        ]
        return element_tables, denormalized_columns

    def _table_relationship(
        self,
        table1: TableInput,
        table2: TableInput,
    ) -> tuple[Column, Column]:
        """Return columns used to relate two tables."""
        table1 = self.name_to_table(table1)
        table2 = self.name_to_table(table2)
        relationships = [
            (fk.foreign_key_columns[0], fk.referenced_columns[0]) for fk in table1.foreign_keys if fk.pk_table == table2
        ]
        relationships.extend(
            [(fk.referenced_columns[0], fk.foreign_key_columns[0]) for fk in table1.referenced_by if fk.table == table2]
        )
        if len(relationships) != 1:
            raise DerivaMLException(
                f"Ambiguous linkage between {table1.name} and {table2.name}: {[(r[0].name, r[1].name) for r in relationships]}"
            )
        return relationships[0]

    def _schema_to_paths(
        self,
        root: Table | None = None,
        path: list[Table] | None = None,
    ) -> list[list[Table]]:
        """Return a list of paths through the schema graph.

        Args:
            root: The root table to start from.
            path: The current path being built.

        Returns:
            A list of paths through the schema graph.
        """
        path = path or []

        root = root or self.model.schemas[self.ml_schema].tables["Dataset"]
        path = path.copy() if path else []
        parent = path[-1] if path else None  # Table that we are coming from.
        path.append(root)
        paths = [path]

        def find_arcs(table: Table) -> set[Table]:
            """Given a path through the model, return the FKs that link the tables"""
            arc_list = [fk.pk_table for fk in table.foreign_keys] + [fk.table for fk in table.referenced_by]
            arc_list = [t for t in arc_list if t.schema.name in {self.domain_schema, self.ml_schema}]
            domain_tables = [t for t in arc_list if t.schema.name == self.domain_schema]
            if multiple_columns := [c for c, cnt in Counter(domain_tables).items() if cnt > 1]:
                raise DerivaMLException(f"Ambiguous relationship in {table.name} {multiple_columns}")
            return set(arc_list)

        def is_nested_dataset_loopback(n1: Table, n2: Table) -> bool:
            """Test to see if node is an association table used to link elements to datasets."""
            # If we have node_name <- node_name_dataset-> Dataset then we are looping
            # back around to a new dataset element
            dataset_table = self.model.schemas[self.ml_schema].tables["Dataset"]
            assoc_table = [a for a in dataset_table.find_associations() if a.table == n2]
            return len(assoc_table) == 1 and n1 != dataset_table

        # Don't follow vocabulary terms back to their use.
        if self.is_vocabulary(root):
            return paths

        for child in find_arcs(root):
            if child.name in {"Dataset_Execution", "Dataset_Dataset", "Execution"}:
                continue
            if child == parent:
                # Don't loop back via referred_by
                continue
            if is_nested_dataset_loopback(root, child):
                continue
            if child in path:
                raise DerivaMLException(f"Cycle in schema path: {child.name} path:{[p.name for p in path]}")

            paths.extend(self._schema_to_paths(child, path))
        return paths

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def create_table(self, table_def: TableDefinition) -> Table:
        """Create a new table from TableDefinition."""
        return self.model.schemas[self.domain_schema].create_table(table_def.model_dump())



import logging
from pathlib import Path
from typing import Any

from hydra.conf import HydraConf, RunDir
from hydra.core.hydra_config import HydraConfig
from hydra_zen import store
from omegaconf import OmegaConf
from pydantic import BaseModel, model_validator

from deriva_ml.core.definitions import ML_SCHEMA


class DerivaMLConfig(BaseModel):
    hostname: str
    catalog_id: str | int = 1
    domain_schema: str | None = None
    project_name: str | None = None
    cache_dir: str | Path | None = None
    working_dir: str | Path | None = None
    hydra_runtime_output_dir: str | Path | None = None
    ml_schema: str = ML_SCHEMA
    logging_level: Any = logging.WARNING
    deriva_logging_level: Any = logging.WARNING
    credential: Any = None
    use_minid: bool = True
    check_auth: bool = True

    @model_validator(mode="after")
    def init_working_dir(self):
        """
        Sets up the working directory for the model.

        This method configures the working directory, ensuring that all required
        file operations are performed in the appropriate location. If the user does not
        specify a directory, a default directory based on the user's home directory
        or username will be used.

        This is a repeat of what is in the DerivaML.__init__ bu we put this here so that the working
        directory is available to hydra.

        Returns:
            Self: The object instance with the working directory initialized.
        """

        self.working_dir = DerivaMLConfig.compute_workdir(self.working_dir)
        self.hydra_runtime_output_dir = Path(HydraConfig.get().runtime.output_dir)
        return self

    @staticmethod
    def compute_workdir(working_dir) -> Path:
        # Create a default working directory if none is provided
        working_dir = Path(working_dir) if working_dir else Path.home() / "deriva-ml"
        return working_dir.absolute()


OmegaConf.register_new_resolver("compute_workdir", DerivaMLConfig.compute_workdir, replace=True)
store(
    HydraConf(
        run=RunDir("${compute_workdir:${deriva_ml.working_dir}}/hydra/${now:%Y-%m-%d_%H-%M-%S}"),
        output_subdir="hydra-config",
    ),
    group="hydra",
    name="config",
)

store.add_to_hydra_store()



"""Enumeration classes for DerivaML.

This module provides enumeration classes used throughout DerivaML for representing states, statuses,
types, and vocabularies. Each enum class represents a specific set of constants used in the system.

Classes:
    BaseStrEnum: Base class for string-based enums.
    UploadState: States for file upload operations.
    Status: Execution status values.
    BuiltinTypes: ERMrest built-in data types.
    MLVocab: Controlled vocabulary types.
    MLAsset: Asset type identifiers.
    ExecMetadataType: Execution metadata type identifiers.
    ExecAssetType: Execution asset type identifiers.
"""

from enum import Enum

from deriva.core.ermrest_model import builtin_types


class BaseStrEnum(str, Enum):
    """Base class for string-based enumerations.

    Extends both str and Enum to create string enums that are both string-like and enumerated.
    This provides type safety while maintaining string compatibility.

    Example:
        >>> class MyEnum(BaseStrEnum):
        ...     VALUE = "value"
        >>> isinstance(MyEnum.VALUE, str)  # True
        >>> isinstance(MyEnum.VALUE, Enum)  # True
    """

    pass


class UploadState(Enum):
    """File upload operation states.

    Represents the various states a file upload operation can be in, from initiation to completion.

    Attributes:
        success (int): Upload completed successfully.
        failed (int): Upload failed.
        pending (int): Upload is queued.
        running (int): Upload is in progress.
        paused (int): Upload is temporarily paused.
        aborted (int): Upload was aborted.
        cancelled (int): Upload was cancelled.
        timeout (int): Upload timed out.
    """

    success = 0
    failed = 1
    pending = 2
    running = 3
    paused = 4
    aborted = 5
    cancelled = 6
    timeout = 7


class Status(BaseStrEnum):
    """Execution status values.

    Represents the various states an execution can be in throughout its lifecycle.

    Attributes:
        initializing (str): Initial setup is in progress.
        created (str): Execution record has been created.
        pending (str): Execution is queued.
        running (str): Execution is in progress.
        aborted (str): Execution was manually stopped.
        completed (str): Execution finished successfully.
        failed (str): Execution encountered an error.
    """

    initializing = "Initializing"
    created = "Created"
    pending = "Pending"
    running = "Running"
    aborted = "Aborted"
    completed = "Completed"
    failed = "Failed"


class BuiltinTypes(Enum):
    """ERMrest built-in data types.

    Maps ERMrest's built-in data types to their type names. These types are used for defining
    column types in tables and for type validation.

    Attributes:
        text (str): Text/string type.
        int2 (str): 16-bit integer.
        jsonb (str): Binary JSON.
        float8 (str): 64-bit float.
        timestamp (str): Timestamp without timezone.
        int8 (str): 64-bit integer.
        boolean (str): Boolean type.
        json (str): JSON type.
        float4 (str): 32-bit float.
        int4 (str): 32-bit integer.
        timestamptz (str): Timestamp with timezone.
        date (str): Date type.
        ermrest_rid (str): Resource identifier.
        ermrest_rcb (str): Record created by.
        ermrest_rmb (str): Record modified by.
        ermrest_rct (str): Record creation time.
        ermrest_rmt (str): Record modification time.
        markdown (str): Markdown text.
        longtext (str): Long text.
        ermrest_curie (str): Compact URI.
        ermrest_uri (str): URI type.
        color_rgb_hex (str): RGB color in hex.
        serial2 (str): 16-bit auto-incrementing.
        serial4 (str): 32-bit auto-incrementing.
        serial8 (str): 64-bit auto-incrementing.
    """

    text = builtin_types.text.typename
    int2 = builtin_types.int2.typename
    jsonb = builtin_types.json.typename
    float8 = builtin_types.float8.typename
    timestamp = builtin_types.timestamp.typename
    int8 = builtin_types.int8.typename
    boolean = builtin_types.boolean.typename
    json = builtin_types.json.typename
    float4 = builtin_types.float4.typename
    int4 = builtin_types.int4.typename
    timestamptz = builtin_types.timestamptz.typename
    date = builtin_types.date.typename
    ermrest_rid = builtin_types.ermrest_rid.typename
    ermrest_rcb = builtin_types.ermrest_rcb.typename
    ermrest_rmb = builtin_types.ermrest_rmb.typename
    ermrest_rct = builtin_types.ermrest_rct.typename
    ermrest_rmt = builtin_types.ermrest_rmt.typename
    markdown = builtin_types.markdown.typename
    longtext = builtin_types.longtext.typename
    ermrest_curie = builtin_types.ermrest_curie.typename
    ermrest_uri = builtin_types.ermrest_uri.typename
    color_rgb_hex = builtin_types.color_rgb_hex.typename
    serial2 = builtin_types.serial2.typename
    serial4 = builtin_types.serial4.typename
    serial8 = builtin_types.serial8.typename


class MLVocab(BaseStrEnum):
    """Controlled vocabulary type identifiers.

    Defines the names of controlled vocabulary tables used in DerivaML for various types
    of entities and attributes.

    Attributes:
        dataset_type (str): Dataset classification vocabulary.
        workflow_type (str): Workflow classification vocabulary.
        asset_type (str): Asset classification vocabulary.
        asset_role (str): Asset role classification vocabulary.
    """

    dataset_type = "Dataset_Type"
    workflow_type = "Workflow_Type"
    asset_type = "Asset_Type"
    asset_role = "Asset_Role"
    feature_name = "Feature_Name"


class MLAsset(BaseStrEnum):
    """Asset type identifiers.

    Defines the types of assets that can be associated with executions.

    Attributes:
        execution_metadata (str): Metadata about an execution.
        execution_asset (str): Asset produced by an execution.
    """

    execution_metadata = "Execution_Metadata"
    execution_asset = "Execution_Asset"


class MLTable(BaseStrEnum):
    dataset = "Dataset"
    workflow = "Workflow"
    file = "File"
    asset = "Asset"
    execution = "Execution"
    dataset_version = "Dataset_Version"
    execution_metadata = "Execution_Metadata"
    execution_asset = "Execution_Asset"


class ExecMetadataType(BaseStrEnum):
    """Execution metadata type identifiers.

    Defines the types of metadata that can be associated with an execution.

    Attributes:
        execution_config (str): Execution configuration data.
        runtime_env (str): Runtime environment information.
    """

    execution_config = "Execution_Config"
    runtime_env = "Runtime_Env"


class ExecAssetType(BaseStrEnum):
    """Execution asset type identifiers.

    Defines the types of assets that can be produced during an execution.

    Attributes:
        input_file (str): Input file used by the execution.
        output_file (str): Output file produced by the execution.
        notebook_output (str): Jupyter notebook output from the execution.
    """

    input_file = "Input_File"
    output_file = "Output_File"
    notebook_output = "Notebook_Output"
    model_file = "Model_File"



from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING

# Safe imports - no circular dependencies
from deriva_ml.core.config import DerivaMLConfig
from deriva_ml.core.definitions import (
    RID,
    BuiltinTypes,
    ColumnDefinition,
    DerivaAssetColumns,
    DerivaSystemColumns,
    ExecAssetType,
    ExecMetadataType,
    FileSpec,
    FileUploadState,
    ForeignKeyDefinition,
    KeyDefinition,
    MLAsset,
    MLVocab,
    TableDefinition,
    UploadState,
)
from deriva_ml.core.exceptions import (
    DerivaMLException,
    DerivaMLInvalidTerm,
    DerivaMLTableTypeError,
)

# Type-checking only - avoid circular import at runtime
if TYPE_CHECKING:
    from deriva_ml.core.base import DerivaML


# Lazy import function for runtime usage
def __getattr__(name):
    """Lazy import to avoid circular dependencies."""
    if name == "DerivaML":
        from deriva_ml.core.base import DerivaML

        return DerivaML
    elif name == "Execution":
        from deriva_ml.execution.execution import Execution

        return Execution
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "DerivaML",  # Lazy-loaded
    "DerivaMLConfig",
    # Exceptions
    "DerivaMLException",
    "DerivaMLInvalidTerm",
    "DerivaMLTableTypeError",
    # Definitions
    "RID",
    "BuiltinTypes",
    "ColumnDefinition",
    "DerivaSystemColumns",
    "DerivaAssetColumns",
    "ExecAssetType",
    "ExecMetadataType",
    "FileSpec",
    "FileUploadState",
    "ForeignKeyDefinition",
    "KeyDefinition",
    "MLAsset",
    "MLVocab",
    "TableDefinition",
    "UploadState",
]

try:
    __version__ = version("deriva_ml")
except PackageNotFoundError:
    # package is not installed
    pass



#!/usr/bin/env python3
"""
Release helper: seed initial semver tag if none exists, otherwise bump with bump-my-version.

Usage:
  python release.py [patch|minor|major]

Env vars:
  START   - initial version if no tag exists (default: 0.1.0)
  PREFIX  - tag prefix (default: v)
"""

from __future__ import annotations

import argparse
import os
import shutil
import subprocess
import sys
from typing import Sequence


def run(
    cmd: Sequence[str], check: bool = True, capture: bool = False, quiet: bool = False
) -> subprocess.CompletedProcess:
    if not quiet:
        print(f"$ {' '.join(cmd)}")
    return subprocess.run(
        cmd,
        check=check,
        capture_output=capture,
        text=True,
    )


def in_git_repo() -> bool:
    try:
        run(["git", "rev-parse", "--is-inside-work-tree"], capture=True)
        return True
    except subprocess.CalledProcessError:
        return False


def has_commits() -> bool:
    try:
        run(["git", "log", "-1"], capture=True)
        return True
    except subprocess.CalledProcessError:
        return False


def latest_semver_tag(prefix: str) -> str | None:
    # Use git's matcher to keep parity with Bash: prefix + x.y.z
    pattern = f"{prefix}[0-9]*.[0-9]*.[0-9]*"
    try:
        cp = run(["git", "describe", "--tags", "--abbrev=0", "--match", pattern], capture=True, quiet=True)
        tag = cp.stdout.strip()
        return tag or None
    except subprocess.CalledProcessError:
        return None


def seed_initial_tag(tag: str) -> None:
    print(f"No existing semver tag found. Seeding initial tag: {tag}")
    run(["git", "tag", tag, "-m", f"Initial release {tag}"])
    # Push tags (ignore failure to keep parity with bash's simple flow)
    run(["git", "push", "--tags"])


def require_tool(name: str) -> None:
    if shutil.which(name) is None:
        print(f"Error: required tool '{name}' not found on PATH.", file=sys.stderr)
        sys.exit(1)


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Set a new version tag for the current repository, and push to remote."
    )
    parser.add_argument(
        "bump", nargs="?", default="patch", choices=["patch", "minor", "major"], help="Which semver part to bump."
    )
    args = parser.parse_args()

    start = os.environ.get("START", "0.1.0")
    prefix = os.environ.get("PREFIX", "v")

    # Sanity checks
    require_tool("git")
    require_tool("uv")

    if not in_git_repo():
        print("Not a git repo.", file=sys.stderr)
        return 1

    # Ensure tags visible in shallow clones
    try:
        run(["git", "fetch", "--tags", "--quiet"], check=False)
    except Exception:
        pass  # non-fatal

    if not has_commits():
        print("No commits found. Commit something before tagging.", file=sys.stderr)
        return 1

    # Find latest semver tag with prefix
    tag = latest_semver_tag(prefix)
    print(f"Latest semver tag: {tag}")
    if not tag:
        seed_initial_tag(f"{prefix}{start}")
        print(f"Seeded {prefix}{start}. Done.")
        return 0

    print(f"Bumping version: {args.bump}")

    # Bump using bump-my-version via uv
    # Mirrors: uv run bump-my-version bump $BUMP --verbose
    try:
        run(["uv", "run", "bump-my-version", "bump", args.bump, "--verbose"])
    except subprocess.CalledProcessError as e:
        print(e.stdout or "", end="")
        print(e.stderr or "", end="", file=sys.stderr)
        return e.returncode

    # Push commits and tags
    print("Pushing changes to remote repository...")
    run(["git", "push", "--follow-tags"])

    # Retrieve new version tag
    try:
        cp = run(["git", "describe", "--tags", "--abbrev=0"], capture=True)
        new_tag = cp.stdout.strip()
        print(f"New version tag: {new_tag}")
    except subprocess.CalledProcessError:
        print("Warning: unable to determine new tag via git describe.", file=sys.stderr)

    print("Release process complete!")
    return 0


if __name__ == "__main__":
    sys.exit(main())



from typing import TYPE_CHECKING

# Safe imports - no circular dependencies
from deriva_ml.execution.execution_configuration import ExecutionConfiguration, AssetRIDConfig
from deriva_ml.execution.workflow import Workflow

if TYPE_CHECKING:
    from deriva_ml.execution.execution import Execution


# Lazy import for runtime
def __getattr__(name):
    """Lazy import to avoid circular dependencies."""
    if name == "Execution":
        from deriva_ml.execution.execution import Execution

        return Execution
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "Execution",  # Lazy-loaded
    "ExecutionConfiguration",
    "Workflow",
    "AssetRIDConfig"
]



"""
Constants used throughout the DerivaML package.
"""

from __future__ import annotations

from typing import NewType, TypeAlias

from pydantic import constr

# Schema name
ML_SCHEMA = "deriva-ml"

# Special RID for dry runs
DRY_RUN_RID = "0000"

# Regular expression parts for RIDs
rid_part = r"(?P<rid>(?:[A-Z\d]{1,4}|[A-Z\d]{1,4}(?:-[A-Z\d]{4})+))"
snapshot_part = r"(?:@(?P<snapshot>(?:[A-Z\d]{1,4}|[A-Z\d]{1,4}(?:-[A-Z\d]{4})+)))?"
rid_regex = f"^{rid_part}{snapshot_part}$"

# RID type definition
BaseRIDString = constr(pattern=rid_regex)
# RID = TypeVar("RID", bound=BaseRIDString)
RIDType: TypeAlias = constr(pattern=rid_regex)
RID = NewType("RID", BaseRIDString)

# System columns in Deriva
DerivaSystemColumns = ["RID", "RCT", "RMT", "RCB", "RMB"]
DerivaAssetColumns = {
    "Filename",
    "URL",
    "Length",
    "MD5",
    "Description",
}.union(set(DerivaSystemColumns))



from __future__ import annotations

import atexit
import itertools
import logging
import string
from collections.abc import Iterator, Sequence
from datetime import datetime
from numbers import Integral
from pathlib import Path
from random import choice, randint, random
from tempfile import TemporaryDirectory

from deriva.core import ErmrestCatalog
from deriva.core.ermrest_model import Column, Schema, Table, builtin_types
from pydantic import BaseModel, ConfigDict
from requests.exceptions import HTTPError

from deriva_ml import DerivaML, MLVocab
from deriva_ml.core.definitions import RID, BuiltinTypes, ColumnDefinition
from deriva_ml.dataset.aux_classes import DatasetVersion
from deriva_ml.execution.execution import Execution
from deriva_ml.execution.execution_configuration import ExecutionConfiguration
from deriva_ml.schema import (
    create_ml_catalog,
)
from deriva_ml.schema.annotations import catalog_annotation

try:
    from icecream import ic

    ic.configureOutput(includeContext=True)
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


TEST_DATASET_SIZE = 12


def populate_demo_catalog(ml_instance: DerivaML) -> None:
    # Delete any vocabularies and features.
    domain_schema = ml_instance.pathBuilder.schemas[ml_instance.domain_schema]
    subject = domain_schema.tables["Subject"]
    ss = subject.insert([{"Name": f"Thing{t + 1}"} for t in range(TEST_DATASET_SIZE)])

    ml_instance.add_term(
        MLVocab.workflow_type,
        "Demo Catalog Creation",
        description="A workflow demonstrating how to create a demo catalog.",
    )
    execution = ml_instance.create_execution(
        ExecutionConfiguration(
            workflow=ml_instance.create_workflow(name="Demo Catalog", workflow_type="Demo Catalog Creation")
        )
    )
    with execution.execute() as e:
        for s in ss:
            image_file = e.asset_file_path(
                "Image",
                f"test_{s['RID']}.txt",
                Subject=s["RID"],
                Acquisition_Time=datetime.now(),
                Acquisition_Date=datetime.now().date(),
            )
            with image_file.open("w") as f:
                f.write(f"Hello there {random()}\n")
        execution.upload_execution_outputs()


class DatasetDescription(BaseModel):
    types: list[str]  # Types of the dataset.
    description: str  # Description.
    members: dict[
        str, int | list[DatasetDescription]
    ]  # Either a list of nested dataset, or then number of elements to add
    member_rids: dict[str, list[RID]] = {}  # The rids of the members of the dataset.
    version: DatasetVersion = DatasetVersion(1, 0, 0)  # The initial version.
    rid: RID = None  # RID of dataset that was created.

    model_config = ConfigDict(arbitrary_types_allowed=True)


def create_datasets(
    client: Execution,
    spec: DatasetDescription,
    member_rids: dict[str, Iterator[RID]],
) -> DatasetDescription:
    """
    Create a dataset per `spec`, then add child members (either by slicing
    off pre-generated RIDs or by recursing on nested specs).
    """
    dataset_rid = client.create_dataset(
        dataset_types=spec.types,
        description=spec.description,
        version=spec.version,
    )

    result_spec = DatasetDescription(
        description=spec.description,
        members={},
        types=spec.types,
        rid=dataset_rid,
        version=spec.version,
    )
    dataset_rids = {}
    for member_type, value in spec.members.items():
        if isinstance(value, Sequence) and not isinstance(value, (str, bytes)):
            nested_specs: list[DatasetDescription] = list(value)
            rids: list[RID] = []
            for child_spec in nested_specs:
                child_ds = create_datasets(client, child_spec, member_rids)
                result_spec.members.setdefault(member_type, []).append(child_ds)
                rids.append(child_ds.rid)
        elif isinstance(value, Integral):
            count = int(value)
            # take exactly `count` RIDs (or an empty list if count <= 0)
            rids = list(itertools.islice(member_rids[member_type], count))
            assert len(rids) == count, f"Expected {count} RIDs, got {len(rids)}"
            result_spec.members[member_type] = count
        else:
            raise TypeError(
                f"Expected spec.members['{member_type}'] to be either an int or a list, got {type(value).__name__!r}"
            )

        # attach and record
        if rids:
            dataset_rids[member_type] = rids
            result_spec.member_rids.setdefault(member_type, []).extend(rids)
    client.add_dataset_members(dataset_rid, dataset_rids, description="Added by create_datasets")

    return result_spec


def dataset_spec() -> DatasetDescription:
    dataset = DatasetDescription(
        description="A dataset",
        members={"Subject": 2},
        types=[],
    )

    training_dataset = DatasetDescription(
        description="A dataset that is nested",
        members={"Dataset": [dataset, dataset], "Image": 2},
        types=["Testing"],
    )

    testing_dataset = DatasetDescription(
        description="A dataset that is nested",
        members={"Dataset": [dataset, dataset], "Image": 2},
        types=["Testing"],
    )

    double_nested_dataset = DatasetDescription(
        description="A dataset that is double nested",
        members={"Dataset": [training_dataset, testing_dataset]},
        types=["Complete"],
    )
    return double_nested_dataset


def create_demo_datasets(ml_instance: DerivaML) -> DatasetDescription:
    """Create datasets from a populated catalog."""
    ml_instance.add_dataset_element_type("Subject")
    ml_instance.add_dataset_element_type("Image")

    _type_rid = ml_instance.add_term("Dataset_Type", "Complete", synonyms=["Whole"], description="A test")
    _training_rid = ml_instance.add_term("Dataset_Type", "Training", synonyms=["Train"], description="A training set")
    _testing_rid = ml_instance.add_term("Dataset_Type", "Testing", description="A testing set")

    table_path = ml_instance.catalog.getPathBuilder().schemas[ml_instance.domain_schema].tables["Subject"]
    subject_rids = [i["RID"] for i in table_path.entities().fetch()]
    table_path = ml_instance.catalog.getPathBuilder().schemas[ml_instance.domain_schema].tables["Image"]
    image_rids = [i["RID"] for i in table_path.entities().fetch()]

    ml_instance.add_term(
        MLVocab.workflow_type,
        "Create Dataset Workflow",
        description="A Workflow that creates a new dataset.",
    )
    dataset_workflow = ml_instance.create_workflow(name="API Workflow", workflow_type="Create Dataset Workflow")

    dataset_execution = ml_instance.create_execution(
        ExecutionConfiguration(workflow=dataset_workflow, description="Create Dataset")
    )

    with dataset_execution.execute() as exe:
        spec = dataset_spec()
        dataset = create_datasets(exe, spec, {"Subject": iter(subject_rids), "Image": iter(image_rids)})
    return dataset


def create_demo_features(ml_instance: DerivaML) -> None:
    ml_instance.create_vocabulary("SubjectHealth", "A vocab")
    ml_instance.add_term(
        "SubjectHealth",
        "Sick",
        description="The subject self reports that they are sick",
    )
    ml_instance.add_term(
        "SubjectHealth",
        "Well",
        description="The subject self reports that they feel well",
    )
    ml_instance.create_vocabulary("ImageQuality", "Controlled vocabulary for image quality")
    ml_instance.add_term("ImageQuality", "Good", description="The image is good")
    ml_instance.add_term("ImageQuality", "Bad", description="The image is bad")
    box_asset = ml_instance.create_asset("BoundingBox", comment="A file that contains a cropped version of a image")

    ml_instance.create_feature(
        "Subject",
        "Health",
        terms=["SubjectHealth"],
        metadata=[ColumnDefinition(name="Scale", type=BuiltinTypes.int2, nullok=True)],
        optional=["Scale"],
    )
    ml_instance.create_feature("Image", "BoundingBox", assets=[box_asset])
    ml_instance.create_feature("Image", "Quality", terms=["ImageQuality"])

    ImageQualityFeature = ml_instance.feature_record_class("Image", "Quality")
    ImageBoundingboxFeature = ml_instance.feature_record_class("Image", "BoundingBox")
    SubjectWellnessFeature = ml_instance.feature_record_class("Subject", "Health")

    # Get the workflow for this notebook

    ml_instance.add_term(
        MLVocab.workflow_type,
        "Feature Notebook Workflow",
        description="A Workflow that uses Deriva ML API",
    )
    ml_instance.add_term(MLVocab.asset_type, "API_Model", description="Model for our Notebook workflow")
    notebook_workflow = ml_instance.create_workflow(name="API Workflow", workflow_type="Feature Notebook Workflow")

    feature_execution = ml_instance.create_execution(
        ExecutionConfiguration(workflow=notebook_workflow, description="Our Sample Workflow instance")
    )

    subject_rids = [i["RID"] for i in ml_instance.domain_path.tables["Subject"].entities().fetch()]
    image_rids = [i["RID"] for i in ml_instance.domain_path.tables["Image"].entities().fetch()]
    _subject_feature_list = [
        SubjectWellnessFeature(
            Subject=subject_rid,
            Execution=feature_execution.execution_rid,
            SubjectHealth=choice(["Well", "Sick"]),
            Scale=randint(1, 10),
        )
        for subject_rid in subject_rids
    ]

    # Create a new set of images.  For fun, lets wrap this in an execution so we get status updates
    bounding_box_files = []
    for i in range(10):
        bounding_box_file = feature_execution.asset_file_path("BoundingBox", f"box{i}.txt")
        with bounding_box_file.open("w") as fp:
            fp.write(f"Hi there {i}")
        bounding_box_files.append(bounding_box_file)

    image_bounding_box_feature_list = [
        ImageBoundingboxFeature(
            Image=image_rid,
            BoundingBox=asset_name,
        )
        for image_rid, asset_name in zip(image_rids, itertools.cycle(bounding_box_files))
    ]

    image_quality_feature_list = [
        ImageQualityFeature(
            Image=image_rid,
            ImageQuality=choice(["Good", "Bad"]),
        )
        for image_rid in image_rids
    ]

    subject_feature_list = [
        SubjectWellnessFeature(
            Subject=subject_rid,
            SubjectHealth=choice(["Well", "Sick"]),
            Scale=randint(1, 10),
        )
        for subject_rid in subject_rids
    ]

    with feature_execution.execute() as execution:
        execution.add_features(image_bounding_box_feature_list)
        execution.add_features(image_quality_feature_list)
        execution.add_features(subject_feature_list)

    feature_execution.upload_execution_outputs()


def create_demo_files(ml_instance: DerivaML):
    """Create demo files for testing purposes.

    Args:
        ml_instance: The DerivaML instance to create files for.

    Returns:
        None. Creates files in the working directory.
    """

    def random_string(length: int) -> str:
        """Generate a random string of specified length.

        Args:
            length: The length of the string to generate.

        Returns:
            A random string of the specified length.
        """
        return "".join(random.choice(string.ascii_letters) for _ in range(length))

    test_dir = ml_instance.working_dir / "test_dir"
    test_dir.mkdir(parents=True, exist_ok=True)
    d1 = test_dir / "d1"
    d1.mkdir(parents=True, exist_ok=True)
    d2 = test_dir / "d2"
    d2.mkdir(parents=True, exist_ok=True)

    # Create some demo files
    for d in [test_dir, d1, d2]:
        for i in range(5):
            fname = Path(d) / f"file{i}.{random.choice(['txt', 'jpeg'])}"
            with fname.open("w") as f:
                f.write(random_string(10))
    ml_instance.add_term(MLVocab.workflow_type, "File Test Workflow", description="Test workflow")


def create_domain_schema(catalog: ErmrestCatalog, sname: str) -> None:
    """
    Create a domain schema.  Assumes that the ml-schema has already been created.
    :param sname:
    :return:
    """
    model = catalog.getCatalogModel()
    _ = model.schemas["deriva-ml"]

    try:
        model.schemas[sname].drop(cascade=True)
    except KeyError:
        pass
    except HTTPError as e:
        print(e)
        if f"Schema {sname} does not exist" in str(e):
            pass
        else:
            raise e

    domain_schema = model.create_schema(Schema.define(sname, annotations={"name_style": {"underline_space": True}}))
    subject_table = domain_schema.create_table(
        Table.define("Subject", column_defs=[Column.define("Name", builtin_types.text)])
    )
    with TemporaryDirectory() as tmpdir:
        ml_instance = DerivaML(hostname=catalog.deriva_server.server, catalog_id=catalog.catalog_id, working_dir=tmpdir)
        ml_instance.create_asset(
            "Image",
            column_defs=[
                Column.define("Acquisition_Time", builtin_types.timestamp),
                Column.define("Acquisition_Date", builtin_types.date),
            ],
            referenced_tables=[subject_table],
        )
        catalog_annotation(ml_instance.model)


def destroy_demo_catalog(catalog):
    """Destroy the demo catalog and clean up resources.

    Args:
        catalog: The ErmrestCatalog instance to destroy.

    Returns:
        None. Destroys the catalog.
    """
    catalog.delete_ermrest_catalog(really=True)


def create_demo_catalog(
    hostname,
    domain_schema="demo-schema",
    project_name="ml-test",
    populate=True,
    create_features=False,
    create_datasets=False,
    on_exit_delete=True,
    logging_level=logging.WARNING,
) -> ErmrestCatalog:
    test_catalog = create_ml_catalog(hostname, project_name=project_name)
    if on_exit_delete:
        atexit.register(destroy_demo_catalog, test_catalog)
    try:
        with TemporaryDirectory() as tmpdir:
            create_domain_schema(test_catalog, domain_schema)
            ml_instance = DerivaML(
                hostname,
                catalog_id=test_catalog.catalog_id,
                domain_schema=domain_schema,
                working_dir=tmpdir,
                logging_level=logging_level,
            )

            if populate or create_features or create_datasets:
                populate_demo_catalog(ml_instance)
                if create_features:
                    create_demo_features(ml_instance)
                if create_datasets:
                    create_demo_datasets(ml_instance)

    except Exception:
        # on failure, delete catalog and re-raise exception
        test_catalog.delete_ermrest_catalog(really=True)
        raise
    return test_catalog


class DemoML(DerivaML):
    def __init__(
        self,
        hostname,
        catalog_id,
        cache_dir: str | None = None,
        working_dir: str | None = None,
        use_minid=True,
    ):
        super().__init__(
            hostname=hostname,
            catalog_id=catalog_id,
            project_name="ml-test",
            cache_dir=cache_dir,
            working_dir=working_dir,
            use_minid=use_minid,
        )



"""Module to run a notebook using papermill"""

import json
import os
import tempfile
from pathlib import Path

import nbformat
import papermill as pm
import yaml
from deriva.core import BaseCLI
from jupyter_client.kernelspec import KernelSpecManager
from nbconvert import MarkdownExporter

from deriva_ml import DerivaML, ExecAssetType, Execution, ExecutionConfiguration, MLAsset, Workflow


class DerivaMLRunNotebookCLI(BaseCLI):
    """Main class to part command line arguments and call model"""

    def __init__(self, description, epilog, **kwargs):
        BaseCLI.__init__(self, description, epilog, **kwargs)
        Workflow._check_nbstrip_status()
        self.parser.add_argument("notebook_file", type=Path, help="Path to the notebook file")

        self.parser.add_argument(
            "--file",
            "-f",
            type=Path,
            default=None,
            help="JSON or YAML file with parameter values to inject into the notebook.",
        )

        self.parser.add_argument(
            "--inspect",
            action="store_true",
            help="Display parameters information for the given notebook path.",
        )

        self.parser.add_argument(
            "--log-output",
            action="store_true",
            help="Display logging output from notebook.",
        )

        self.parser.add_argument(
            "--parameter",
            "-p",
            nargs=2,
            action="append",
            metavar=("KEY", "VALUE"),
            default=[],
            help="Provide a parameter name and value to inject into the notebook.",
        )

        self.parser.add_argument(
            "--kernel",
            "-k",
            type=str,
            help="Name of kernel to run..",
            default=self._find_kernel_for_venv(),
        )

    @staticmethod
    def _coerce_number(val: str):
        """
        Try to convert a string to int, then float; otherwise return str.
        """
        try:
            return int(val)
        except ValueError:
            try:
                return float(val)
            except ValueError:
                return val

    def main(self):
        """Parse arguments and set up execution environment."""
        args = self.parse_cli()
        notebook_file: Path = args.notebook_file
        parameter_file = args.file

        # args.parameter is now a list of [KEY, VALUE] lists
        # e.g. [['timeout', '30'], ['name', 'Alice'], ...]
        parameters = {key: self._coerce_number(val) for key, val in args.parameter}

        if parameter_file:
            with parameter_file.open("r") as f:
                if parameter_file.suffix == ".json":
                    parameters |= json.load(f)
                elif parameter_file.suffix == ".yaml":
                    parameters |= yaml.safe_load(f)
                else:
                    print("Parameter file must be an json or YAML file.")
                    exit(1)

        if not (notebook_file.is_file() and notebook_file.suffix == ".ipynb"):
            print(f"Notebook file must be an ipynb file: {notebook_file.name}.")
            exit(1)

        # Create a workflow instance for this specific version of the script.
        # Return an existing workflow if one is found.
        notebook_parameters = pm.inspect_notebook(notebook_file)

        if args.inspect:
            for param, value in notebook_parameters.items():
                print(f"{param}:{value['inferred_type_name']}  (default {value['default']})")
            return
        else:
            notebook_parameters = {k: v["default"] for k, v in notebook_parameters.items()} | parameters
            self.run_notebook(notebook_file.resolve(), parameters, kernel=args.kernel, log=args.log_output)

    @staticmethod
    def _find_kernel_for_venv() -> str | None:
        """
        Return the name and spec of an existing Jupyter kernel corresponding
        to a given Python virtual environment path.

        Parameters
        ----------
        venv_path : str
            Absolute or relative path to the virtual environment.

        Returns
        -------
        dict | None
            The kernel spec (as a dict) if found, or None if not found.
        """
        venv = os.environ.get("VIRTUAL_ENV")
        if not venv:
            return None
        venv_path = Path(venv).resolve()
        ksm = KernelSpecManager()
        for name, spec in ksm.get_all_specs().items():
            kernel_json = spec.get("spec", {})
            argv = kernel_json.get("argv", [])
            # check for python executable path inside argv
            for arg in argv:
                try:
                    if Path(arg).resolve() == venv_path.joinpath("bin", "python").resolve():
                        return name
                except Exception:
                    continue
        return None

    def run_notebook(self, notebook_file: Path, parameters, kernel=None, log=False):
        url, checksum = Workflow.get_url_and_checksum(Path(notebook_file))
        os.environ["DERIVA_ML_WORKFLOW_URL"] = url
        os.environ["DERIVA_ML_WORKFLOW_CHECKSUM"] = checksum
        os.environ["DERIVA_ML_NOTEBOOK_PATH"] = notebook_file.as_posix()
        with tempfile.TemporaryDirectory() as tmpdirname:
            notebook_output = Path(tmpdirname) / Path(notebook_file).name
            execution_rid_path = Path(tmpdirname) / "execution_rid.json"
            os.environ["DERIVA_ML_SAVE_EXECUTION_RID"] = execution_rid_path.as_posix()
            pm.execute_notebook(
                input_path=notebook_file,
                output_path=notebook_output,
                parameters=parameters,
                kernel_name=kernel,
                log_output=log,
            )
            print(f"Notebook output saved to {notebook_output}")
            with execution_rid_path.open("r") as f:
                execution_config = json.load(f)

            if not execution_config:
                print("Execution RID not found.")
                exit(1)

            execution_rid = execution_config["execution_rid"]
            hostname = execution_config["hostname"]
            catalog_id = execution_config["catalog_id"]
            workflow_rid = execution_config["workflow_rid"]
            ml_instance = DerivaML(hostname=hostname, catalog_id=catalog_id, working_dir=tmpdirname)
            workflow_rid = ml_instance.retrieve_rid(execution_config["execution_rid"])["Workflow"]

            execution = Execution(
                configuration=ExecutionConfiguration(workflow=workflow_rid),
                ml_object=ml_instance,
                reload=execution_rid,
            )

            # Generate an HTML version of the output notebook.
            notebook_output_md = notebook_output.with_suffix(".md")
            with notebook_output.open() as f:
                nb = nbformat.read(f, as_version=4)
            # Convert to Markdown
            exporter = MarkdownExporter()
            (body, resources) = exporter.from_notebook_node(nb)

            with notebook_output_md.open("w") as f:
                f.write(body)
            nb = nbformat.read(notebook_output, as_version=4)

            execution.asset_file_path(
                asset_name=MLAsset.execution_asset,
                file_name=notebook_output,
                asset_types=ExecAssetType.notebook_output,
            )

            execution.asset_file_path(
                asset_name=MLAsset.execution_asset,
                file_name=notebook_output_md,
                asset_types=ExecAssetType.notebook_output,
            )
            execution.upload_execution_outputs()

            print(ml_instance.cite(execution_rid))


def main():
    """Main entry point for the notebook runner CLI.

    Creates and runs the DerivaMLRunNotebookCLI instance.

    Returns:
        None. Executes the CLI.
    """
    cli = DerivaMLRunNotebookCLI(description="Deriva ML Execution Script Demo", epilog="")
    cli.main()


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(e)
        exit(1)



"""
Tests for the execution module.
"""

import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory

from deriva_ml import (
    DatasetSpec,
    DerivaML,
    ExecAssetType,
    ExecutionConfiguration,
    MLAsset,
)
from deriva_ml import (
    MLVocab as vc,
)


class TestWorkflow:
    def test_workflow_creation_script(self, test_ml):
        ml_instance = test_ml
        ml_instance.add_term(vc.asset_type, "Test Model", description="Model for our Test workflow")
        ml_instance.add_term(vc.workflow_type, "Test Workflow", description="A ML Workflow that uses Deriva ML API")
        print("Running workflow-test.py ...")
        workflow_script = Path(__file__).parent / "workflow-test.py"

        workflow_table = ml_instance.pathBuilder.schemas[ml_instance.ml_schema].Workflow
        workflows = list(workflow_table.entities().fetch())
        assert 0 == len(workflows)
        result = subprocess.run(
            [
                "python",
                workflow_script.as_posix(),
                ml_instance.catalog.deriva_server.server,
                ml_instance.catalog_id,
            ],
            capture_output=True,
            text=True,
        )
        print(result.stdout)
        print(result.stderr)
        workflows = list(workflow_table.entities().fetch())
        assert 1 == len(workflows)
        workflow_rid = workflows[0]["RID"]
        workflow_url = workflows[0]["URL"]

        workflow_rid = ml_instance.lookup_workflow(workflow_url)
        print(f"Workflow url: {workflow_url}")
        assert workflow_url.endswith("workflow-test.py")

        # Make sure that workflow is not duplicated if created again.
        result = subprocess.run(
            [
                "python",
                workflow_script.as_posix(),
                ml_instance.catalog.deriva_server.server,
                ml_instance.catalog_id,
            ],
            capture_output=True,
            text=True,
        )
        print(result.stdout)
        print(result.stderr)
        new_workflow = result.stdout.strip()
        assert new_workflow == workflow_rid

    def test_workflow_creation_notebook(self, notebook_test):
        ml_instance = notebook_test

        notebook_path = Path(__file__).parent / "workflow-test.ipynb"  # directory where this test lives
        ml_instance.add_term(vc.asset_type, "Test Model", description="Model for our Test workflow")
        ml_instance.add_term(vc.workflow_type, "Test Workflow", description="A ML Workflow that uses Deriva ML API")
        workflow_table = ml_instance.pathBuilder.schemas[ml_instance.ml_schema].Workflow
        workflows = list(workflow_table.entities().fetch())
        assert 0 == len(workflows)

        print("Running notebook...")

        result = subprocess.run(
            [
                "deriva-ml-run-notebook",
                notebook_path.as_posix(),
                "--host",
                ml_instance.catalog.deriva_server.server,
                "--catalog",
                ml_instance.catalog_id,
                "--kernel",
                "test_kernel",
                "--log-output",
            ],
            capture_output=True,
            text=True,
        )
        workflows = list(workflow_table.entities().fetch())
        assert 1 == len(workflows)
        workflow_rid = workflows[0]["RID"]
        workflow_url = workflows[0]["URL"]
        assert workflow_url.endswith("workflow-test.ipynb")

        # Check to make sure that result notebook and HTML version got uploaded.
        execution_assets = ml_instance.list_assets("Execution_Asset")


class TestExecution:
    def test_execution_no_download(self, test_ml):
        ml_instance = test_ml
        ml_instance.add_term(vc.asset_type, "Test Model", description="Model for our Test workflow")
        ml_instance.add_term(vc.workflow_type, "Test Workflow", description="A ML Workflow that uses Deriva ML API")

        api_workflow = ml_instance.create_workflow(
            name="Test Workflow One",
            workflow_type="Test Workflow",
            description="A test operation",
        )

        manual_execution = ml_instance.create_execution(
            ExecutionConfiguration(description="Sample Execution", workflow=api_workflow)
        )

        with manual_execution:
            pass
        manual_execution.upload_execution_outputs()
        assert "Completed" == ml_instance.retrieve_rid(manual_execution.execution_rid)["Status"]

        pb = ml_instance.pathBuilder.schemas[ml_instance.ml_schema]
        execution_metadata_execution = pb.Execution_Metadata_Execution
        execution_metadata = pb.Execution_Metadata
        metadata_path = execution_metadata_execution.path
        metadata_path.filter(execution_metadata_execution.Execution == manual_execution.execution_rid)
        metadata_path.link(execution_metadata)
        execution_metadata = [
            {
                "RID": a["RID"],
                "Filename": a["Filename"],
            }
            for a in metadata_path.entities().fetch()
        ]
        assert "uv.lock" in [d["Filename"] for d in execution_metadata]
        assert "configuration.json" in [d["Filename"] for d in execution_metadata]

        assert 3 == len(execution_metadata)

    def test_execution_upload(self, test_ml):
        ml_instance = test_ml
        ml_instance.add_term(vc.asset_type, "Test Model", description="Model for our Test workflow")
        ml_instance.add_term(vc.workflow_type, "Test Workflow", description="A ML Workflow that uses Deriva ML API")

        api_workflow = ml_instance.create_workflow(
            name="Test Workflow One",
            workflow_type="Test Workflow",
            description="A test operation",
        )

        self.create_execution_asset(ml_instance, api_workflow)

    def test_execution_workflow_download(self, dataset_test, tmp_path):
        """Test creating and configuring an execution workflow."""
        hostname = dataset_test.catalog.hostname
        catalog_id = dataset_test.catalog.catalog_id
        ml_instance = DerivaML(hostname, catalog_id, working_dir=tmp_path, use_minid=False)
        ml_instance.add_term(vc.asset_type, "Test Model", description="Model for our Test workflow")
        ml_instance.add_term(vc.workflow_type, "Test Workflow", description="A ML Workflow that uses Deriva ML API")

        dataset_rid = dataset_test.dataset_description.rid

        # Create a workflow
        api_workflow = ml_instance.create_workflow(
            name="Test Workflow One",
            workflow_type="Test Workflow",
            description="A test operation",
        )

        model_rid = self.create_execution_asset(ml_instance, api_workflow)
        # Create execution configuration
        config = ExecutionConfiguration(
            datasets=[
                DatasetSpec(
                    rid=dataset_rid,
                    version=ml_instance.dataset_version(dataset_rid),
                ),
            ],
            assets=[model_rid],
            description="Sample Execution",
            workflow=api_workflow,
        )

        # Create manual execution
        test_execution = ml_instance.create_execution(config)
        with test_execution.execute() as execution:
            assert execution.asset_paths["Execution_Asset"][0].asset_types[0] == "Model_File"
            assert 1 == len(execution.asset_paths)
            assert 1 == len(execution.datasets)
            assert execution.datasets[0].dataset_rid == dataset_rid

    @staticmethod
    def create_execution_asset(ml_instance: DerivaML, api_workflow):
        manual_execution = ml_instance.create_execution(
            ExecutionConfiguration(description="Sample Execution", workflow=api_workflow)
        )

        with manual_execution.execute() as execution:
            model_file = execution.asset_file_path(
                MLAsset.execution_asset, "API_Model/modelfile.txt", asset_types=ExecAssetType.model_file
            )
            with model_file.open("w") as fp:
                fp.write("My model")
            # Now upload the file and retrieve the RID of the new asset from the returned results.
        uploaded_assets = manual_execution.upload_execution_outputs()
        assert 1 == len(uploaded_assets["deriva-ml/Execution_Asset"])

        pb = ml_instance.pathBuilder.schemas[ml_instance.ml_schema]
        execution_asset_execution = pb.Execution_Asset_Execution
        execution_asset = pb.Execution_Asset
        asset_path = execution_asset_execution.path
        asset_path.filter(execution_asset_execution.Execution == manual_execution.execution_rid)
        asset_path.link(execution_asset)

        execution_asset = [
            {
                "RID": a["RID"],
                "Filename": a["Filename"],
            }
            for a in asset_path.entities().fetch()
        ]

        assert 1 == len(execution_asset)
        with TemporaryDirectory() as tmpdir:
            file = manual_execution.download_asset(execution_asset[0]["RID"], tmpdir, update_catalog=False)
            assert file.name == "modelfile.txt"
        return execution_asset[0]["RID"]



from deriva_ml import (
    DatasetVersion,
    DerivaML,
    VersionPart,
)


class TestDatasetVersion:
    def test_dataset_version_simple(self, test_ml):
        ml_instance = test_ml
        type_rid = ml_instance.add_term("Dataset_Type", "TestSet", description="A test")
        dataset_rid = ml_instance.create_dataset(
            type_rid.name,
            description="A New Dataset",
            version=DatasetVersion(1, 0, 0),
        )
        v0 = ml_instance.dataset_version(dataset_rid)
        assert "1.0.0" == str(v0)
        v1 = ml_instance.increment_dataset_version(dataset_rid=dataset_rid, component=VersionPart.minor)
        assert "1.1.0" == str(v1)

    def test_dataset_version_history(self, test_ml):
        ml_instance = test_ml
        type_rid = ml_instance.add_term("Dataset_Type", "TestSet", description="A test")
        dataset_rid = ml_instance.create_dataset(
            type_rid.name,
            description="A New Dataset",
            version=DatasetVersion(1, 0, 0),
        )
        assert 1 == len(ml_instance.dataset_history(dataset_rid))
        v1 = ml_instance.increment_dataset_version(dataset_rid=dataset_rid, component=VersionPart.minor)
        assert 2 == len(ml_instance.dataset_history(dataset_rid))

    def test_dataset_version(self, dataset_test, tmp_path):
        hostname = dataset_test.catalog.hostname
        catalog_id = dataset_test.catalog.catalog_id
        ml_instance = DerivaML(hostname, catalog_id, working_dir=tmp_path, use_minid=False)
        dataset_description = dataset_test.dataset_description

        nested_datasets = dataset_description.member_rids.get("Dataset", [])
        datasets = [
            dataset
            for nested_description in dataset_description.members.get("Dataset", [])
            for dataset in nested_description.member_rids.get("Dataset", [])
        ]
        _versions = {
            "d0": ml_instance.dataset_version(dataset_description.rid),
            "d1": [ml_instance.dataset_version(v) for v in nested_datasets],
            "d2": [ml_instance.dataset_version(v) for v in datasets],
        }
        ml_instance.increment_dataset_version(nested_datasets[0], VersionPart.major)
        new_versions = {
            "d0": ml_instance.dataset_version(dataset_description.rid),
            "d1": [ml_instance.dataset_version(v) for v in nested_datasets],
            "d2": [ml_instance.dataset_version(v) for v in datasets],
        }

        assert new_versions["d0"].major == 2
        assert new_versions["d2"][0].major == 2



"""
The module implements the sqllite interface to a set of directories representing a dataset bag.
"""

from __future__ import annotations

# Standard library imports
from collections import defaultdict
from copy import copy
from typing import TYPE_CHECKING, Any, Generator, Iterable, cast

import deriva.core.datapath as datapath

# Third-party imports
import pandas as pd

# Local imports
from deriva.core.ermrest_model import Table

# Deriva imports
from pydantic import ConfigDict, validate_call
from sqlalchemy import CompoundSelect, Engine, RowMapping, Select, and_, inspect, select, union
from sqlalchemy.orm import RelationshipProperty, Session
from sqlalchemy.orm.util import AliasedClass

from deriva_ml.core.definitions import RID, VocabularyTerm
from deriva_ml.core.exceptions import DerivaMLException, DerivaMLInvalidTerm
from deriva_ml.feature import Feature

if TYPE_CHECKING:
    from deriva_ml.model.database import DatabaseModel

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


class DatasetBag:
    """
    DatasetBag is a class that manages a materialized bag.  It is created from a locally materialized
    BDBag for a dataset_table, which is created either by DerivaML.create_execution, or directly by
    calling DerivaML.download_dataset.

    A general a bag may contain multiple datasets, if the dataset is nested. The DatasetBag is used to
    represent only one of the datasets in the bag.

    All the metadata associated with the dataset is stored in a SQLLite database that can be queried using SQL.

    Attributes:
        dataset_rid (RID): RID for the specified dataset
        version: The version of the dataset
        model (DatabaseModel): The Database model that has all the catalog metadata associated with this dataset.
            database:
        dbase (sqlite3.Connection): connection to the sqlite database holding table values
        domain_schema (str): Name of the domain schema
    """

    def __init__(self, database_model: DatabaseModel, dataset_rid: RID | None = None) -> None:
        """
        Initialize a DatasetBag instance.

        Args:
            database_model: Database version of the bag.
            dataset_rid: Optional RID for the dataset.
        """
        self.model = database_model
        self.engine = cast(Engine, self.model.engine)
        self.metadata = self.model.metadata

        self.dataset_rid = dataset_rid or self.model.dataset_rid
        if not self.dataset_rid:
            raise DerivaMLException("No dataset RID provided")

        self.model.rid_lookup(self.dataset_rid)  # Check to make sure that this dataset is in the bag.

        self.version = self.model.dataset_version(self.dataset_rid)
        self._dataset_table = self.model.dataset_table

    def __repr__(self) -> str:
        return f"<deriva_ml.DatasetBag object {self.dataset_rid} at {hex(id(self))}>"

    def list_tables(self) -> list[str]:
        """List the names of the tables in the catalog

        Returns:
            A list of table names.  These names are all qualified with the Deriva schema name.
        """
        return self.model.list_tables()

    @staticmethod
    def _find_relationship_attr(source, target):
        """
        Return the relationship attribute (InstrumentedAttribute) on `source`
        that points to `target`. Works with classes or AliasedClass.
        Raises LookupError if not found.
        """
        src_mapper = inspect(source).mapper
        tgt_mapper = inspect(target).mapper

        # collect relationships on the *class* mapper (not on alias)
        candidates: list[RelationshipProperty] = [rel for rel in src_mapper.relationships if rel.mapper is tgt_mapper]

        if not candidates:
            raise LookupError(f"No relationship from {src_mapper.class_.__name__} → {tgt_mapper.class_.__name__}")

        # Prefer MANYTOONE when multiple paths exist (often best for joins)
        candidates.sort(key=lambda r: r.direction.name != "MANYTOONE")
        rel = candidates[0]

        # Bind to the actual source (alias or class)
        return getattr(source, rel.key) if isinstance(source, AliasedClass) else rel.class_attribute

    def _dataset_table_view(self, table: str) -> CompoundSelect[Any]:
        """Return a SQL command that will return all of the elements in the specified table that are associated with
        dataset_rid"""
        table_class = self.model.get_orm_class_by_name(table)
        dataset_table_class = self.model.get_orm_class_by_name(self._dataset_table.name)
        dataset_rids = [self.dataset_rid] + [c.dataset_rid for c in self.list_dataset_children(recurse=True)]

        paths = [[t.name for t in p] for p in self.model._schema_to_paths() if p[-1].name == table]
        sql_cmds = []
        for path in paths:
            path_sql = select(table_class)
            last_class = self.model.get_orm_class_by_name(path[0])
            for t in path[1:]:
                t_class = self.model.get_orm_class_by_name(t)
                path_sql = path_sql.join(self._find_relationship_attr(last_class, t_class))
                last_class = t_class
            path_sql = path_sql.where(dataset_table_class.RID.in_(dataset_rids))
            sql_cmds.append(path_sql)
        return union(*sql_cmds)

    def get_table(self, table: str) -> Generator[tuple, None, None]:
        """Retrieve the contents of the specified table. If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: return: A generator that yields tuples of column values.

        Returns:
          A generator that yields tuples of column values.

        """
        with Session(self.engine) as session:
            result = session.execute(self._dataset_table_view(table))
            for row in result:
                yield row

    def get_table_as_dataframe(self, table: str) -> pd.DataFrame:
        """Retrieve the contents of the specified table as a dataframe.


        If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: Table to retrieve data from.

        Returns:
          A dataframe containing the contents of the specified table.
        """
        return pd.read_sql(self._dataset_table_view(table), self.engine)

    def get_table_as_dict(self, table: str) -> Generator[dict[str, Any], None, None]:
        """Retrieve the contents of the specified table as a dictionary.

        Args:
            table: Table to retrieve data from. f schema is not provided as part of the table name,
                the method will attempt to locate the schema for the table.

        Returns:
          A generator producing dictionaries containing the contents of the specified table as name/value pairs.
        """

        with Session(self.engine) as session:
            result = session.execute(self._dataset_table_view(table))
            for row in result.mappings():
                yield row

    # @validate_call
    def list_dataset_members(self, recurse: bool = False) -> dict[str, list[dict[str, Any]]]:
        """Return a list of entities associated with a specific dataset.

        Args:
           recurse: Whether to include nested datasets.

        Returns:
            Dictionary of entities associated with the dataset.
        """

        # Look at each of the element types that might be in the _dataset_table and get the list of rid for them from
        # the appropriate association table.
        members = defaultdict(list)

        dataset_class = self.model.get_orm_class_for_table(self._dataset_table)
        for element_table in self.model.list_dataset_element_types():
            element_class = self.model.get_orm_class_for_table(element_table)

            assoc_class, dataset_rel, element_rel = self.model.get_orm_association_class(dataset_class, element_class)

            element_table = inspect(element_class).mapped_table
            if element_table.schema != self.model.domain_schema and element_table.name not in ["Dataset", "File"]:
                # Look at domain tables and nested datasets.
                continue
            # Get the names of the columns that we are going to need for linking
            with Session(self.engine) as session:
                sql_cmd = (
                    select(element_class)
                    .join(element_rel)
                    .where(self.dataset_rid == assoc_class.__table__.c["Dataset"])
                )
                # Get back the list of ORM entities and convert them to dictionaries.
                element_entities = session.scalars(sql_cmd).all()
                element_rows = [{c.key: getattr(obj, c.key) for c in obj.__table__.columns} for obj in element_entities]
            members[element_table.name].extend(element_rows)
            if recurse and (element_table.name == self._dataset_table.name):
                # Get the members for all the nested datasets and add to the member list.
                nested_datasets = [d["RID"] for d in element_rows]
                for ds in nested_datasets:
                    nested_dataset = self.model.get_dataset(ds)
                    for k, v in nested_dataset.list_dataset_members(recurse=recurse).items():
                        members[k].extend(v)
        return dict(members)

    def find_features(self, table: str | Table) -> Iterable[Feature]:
        """Find features for a table.

        Args:
            table: The table to find features for.

        Returns:
            An iterable of Feature instances.
        """
        return self.model.find_features(table)

    def list_feature_values(self, table: Table | str, feature_name: str) -> datapath._ResultSet:
        """Return feature values for a table.

        Args:
            table: The table to get feature values for.
            feature_name: Name of the feature.

        Returns:
            Feature values.
        """
        feature = self.model.lookup_feature(table, feature_name)
        feature_class = self.model.get_orm_class_for_table(feature.feature_table)
        with Session(self.engine) as session:
            sql_cmd = select(feature_class)
            return cast(datapath._ResultSet, [row for row in session.execute(sql_cmd).mappings()])

    def list_dataset_element_types(self) -> list[Table]:
        """
        Lists the data types of elements contained within a dataset.

        This method analyzes the dataset and identifies the data types for all
        elements within it. It is useful for understanding the structure and
        content of the dataset and allows for better manipulation and usage of its
        data.

        Returns:
            list[str]: A list of strings where each string represents a data type
            of an element found in the dataset.

        """
        return self.model.list_dataset_element_types()

    def list_dataset_children(self, recurse: bool = False) -> list[DatasetBag]:
        """Get nested datasets.

        Args:
            recurse: Whether to include children of children.

        Returns:
            List of child dataset bags.
        """
        ds_table = self.model.get_orm_class_by_name(f"{self.model.ml_schema}.Dataset")
        nds_table = self.model.get_orm_class_by_name(f"{self.model.ml_schema}.Dataset_Dataset")
        dv_table = self.model.get_orm_class_by_name(f"{self.model.ml_schema}.Dataset_Version")

        with Session(self.engine) as session:
            sql_cmd = (
                select(nds_table.Nested_Dataset, dv_table.Version)
                .join_from(ds_table, nds_table, onclause=ds_table.RID == nds_table.Nested_Dataset)
                .join_from(ds_table, dv_table, onclause=ds_table.Version == dv_table.RID)
                .where(nds_table.Dataset == self.dataset_rid)
            )
            nested = [DatasetBag(self.model, r[0]) for r in session.execute(sql_cmd).all()]

        result = copy(nested)
        if recurse:
            for child in nested:
                result.extend(child.list_dataset_children(recurse))
        return result

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def lookup_term(self, table: str | Table, term_name: str) -> VocabularyTerm:
        """Finds a term in a vocabulary table.

        Searches for a term in the specified vocabulary table, matching either the primary name
        or any of its synonyms.

        Args:
            table: Vocabulary table to search in (name or Table object).
            term_name: Name or synonym of the term to find.

        Returns:
            VocabularyTerm: The matching vocabulary term.

        Raises:
            DerivaMLVocabularyException: If the table is not a vocabulary table, or term is not found.

        Examples:
            Look up by primary name:
                >>> term = ml.lookup_term("tissue_types", "epithelial")
                >>> print(term.description)

            Look up by synonym:
                >>> term = ml.lookup_term("tissue_types", "epithelium")
        """
        # Get and validate vocabulary table reference
        if not self.model.is_vocabulary(table):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")

        # Search for term by name or synonym
        for term in self.get_table_as_dict(table):
            if term_name == term["Name"] or (term["Synonyms"] and term_name in term["Synonyms"]):
                term["Synonyms"] = list(term["Synonyms"])
                return VocabularyTerm.model_validate(term)

        # Term not found
        raise DerivaMLInvalidTerm(table, term_name)

    def _denormalize(self, include_tables: list[str]) -> Select:
        """
        Generates an SQL statement for denormalizing the dataset based on the tables to include. Processes cycles in
        graph relationships, ensures proper join order, and generates selected columns for denormalization.

        Args:
            include_tables (list[str] | None): List of table names to include in the denormalized dataset. If None,
                all tables from the dataset will be included.

        Returns:
            str: SQL query string that represents the process of denormalization.
        """
        # Skip over tables that we don't want to include in the denormalized dataset.
        # Also, strip off the Dataset/Dataset_X part of the path so we don't include dataset columns in the denormalized
        # table.

        def find_relationship(table, join_condition):
            side1 = (join_condition[0].table.name, join_condition[0].name)
            side2 = (join_condition[1].table.name, join_condition[1].name)

            for relationship in inspect(table).relationships:
                local_columns = list(relationship.local_columns)[0].table.name, list(relationship.local_columns)[0].name
                remote_side = list(relationship.remote_side)[0].table.name, list(relationship.remote_side)[0].name
                if local_columns == side1 and remote_side == side2 or local_columns == side2 and remote_side == side1:
                    return relationship
            return None

        join_tables, denormalized_columns = (
            self.model._prepare_wide_table(self, self.dataset_rid, include_tables)
        )

        denormalized_columns = [
            self.model.get_orm_class_by_name(table_name)
            .__table__.columns[column_name]
            .label(f"{table_name}.{column_name}")
            for table_name, column_name in denormalized_columns
        ]
        sql_statements = []
        for key, (path, join_conditions) in join_tables.items():
            sql_statement = select(*denormalized_columns).select_from(
                self.model.get_orm_class_for_table(self._dataset_table)
            )
            for table_name in path[1:]:  # Skip over dataset table
                table_class = self.model.get_orm_class_by_name(table_name)
                on_clause = [
                    getattr(table_class, r.key)
                    for on_condition in join_conditions[table_name]
                    if (r := find_relationship(table_class, on_condition))
                ]
                sql_statement = sql_statement.join(table_class, onclause=and_(*on_clause))
            dataset_rid_list = [self.dataset_rid] + self.list_dataset_children(recurse=True)
            dataset_class = self.model.get_orm_class_by_name(self._dataset_table.name)
            sql_statement = sql_statement.where(dataset_class.RID.in_(dataset_rid_list))
            sql_statements.append(sql_statement)
        return union(*sql_statements)

    def denormalize_as_dataframe(self, include_tables: list[str]) -> pd.DataFrame:
        """
        Denormalize the dataset and return the result as a dataframe.

         This routine will examine the domain schema for the dataset, determine which tables to include and denormalize
        the dataset values into a single wide table.  The result is returned as a generator that returns a dictionary
        for each row in the denormalized wide table.

        The optional argument include_tables can be used to specify a subset of tables to include in the denormalized
        view.  The tables in this argument can appear anywhere in the dataset schema.  The method will determine which
        additional tables are required to complete the denormalization process.  If include_tables is not specified,
        all of the tables in the schema will be included.

        The resulting wide table will include a column for every table needed to complete the denormalization process.

        Args:
            include_tables: List of table names to include in the denormalized dataset.

        Returns:
            Dataframe containing the denormalized dataset.
        """
        return pd.read_sql(self._denormalize(include_tables=include_tables), self.engine)

    def denormalize_as_dict(self, include_tables: list[str]) -> Generator[RowMapping, None, None]:
        """
        Denormalize the dataset and return the result as a set of dictionary's.

        This routine will examine the domain schema for the dataset, determine which tables to include and denormalize
        the dataset values into a single wide table.  The result is returned as a generator that returns a dictionary
        for each row in the denormalized wide table.

        The optional argument include_tables can be used to specify a subset of tables to include in the denormalized
        view.  The tables in this argument can appear anywhere in the dataset schema.  The method will determine which
        additional tables are required to complete the denormalization process.  If include_tables is not specified,
        all of the tables in the schema will be included.

        The resulting wide table will include a only those column for the tables listed in include_columns.

        Args:
            include_tables: List of table names to include in the denormalized dataset. If None, than the entire schema
            is used.

        Returns:
            A generator that returns a dictionary representation of each row in the denormalized dataset.
        """
        with Session(self.engine) as session:
            cursor = session.execute(
                self._denormalize(include_tables=include_tables)
            )
            yield from cursor.mappings()
            for row in cursor.mappings():
                yield row


# Add annotations after definition to deal with forward reference issues in pydantic

DatasetBag.list_dataset_children = validate_call(
    config=ConfigDict(arbitrary_types_allowed=True),
    validate_return=True,
)(DatasetBag.list_dataset_children)



"""A module defining the DatasetLike protocol for dataset operations.

This module contains the definition of the DatasetLike protocol, which
provides an interface for datasets to implement specific functionality related
to listing dataset children. It is particularly useful for ensuring type
compatibility for objects that mimic datasets in their behavior.

Classes:
    DatasetLike: A protocol that specifies methods required for dataset-like
    objects.
"""
from typing import Protocol, runtime_checkable

from deriva_ml.core.definitions import RID


@runtime_checkable
class DatasetLike(Protocol):
    def list_dataset_children(self, dataset_rid: RID, recurse: bool = False) -> list[RID]: ...
